////////////////////////////////////////////////////////////////////////////////
// $RCSfile:$ $Revision:$
// $Source:$
// $Date:$ $Author:$ $State:$
////////////////////////////////////////////////////////////////////////////////
/// \file CIntManager.cpp
/// \brief Definition of CIntManager methods.
///
/// \author Harlen Costa Batagelo (mailto:harlen@dca.fee.unicamp.br)
///
/// See Copyright Notice at the end of the file.
///
/// $Id:$
////////////////////////////////////////////////////////////////////////////////

#include "CIntObj.h"

#if defined(D3D9)
//#define DEBUG_VS ///< Uncomment this line to debug vertex shaders.
//#define DEBUG_PS ///< Uncomment this line to debug pixel shaders. 
#endif

#include <malloc.h>

#if defined(D3D9)
/// Converts a CHAR to TCHAR, allocating the stack memory.
#define C2T(lpa) (((CHAR*)lpa == NULL) ? NULL : \
	ANSIStringToGeneric((TCHAR*) _alloca((strlen(lpa)+1)*2), lpa))
#endif

/// Saturates the x number to minimum value of min.
#define SATURATE_MIN( min, x ) ( ( (x) < (min)) ? (min) : (x) ) 
/// Saturates the x number to maximum value of max.
#define SATURATE_MAX( max, x ) ( ( (x) > (max)) ? (max) : (x) )

#if defined(D3D9)

////////////////////////////////////////////////////////////////////////////////
/// \brief Converts CHAR string to WCHAR.
///
/// UNICODE conversion utility to convert a CHAR string to a WCHAR string.
///
/// \param wstrDestination Resulting WCHAR string 
///        (not allocated by this function).
/// \param strSource Source CHAR string.
/// \param iDestChar Size of wstrDestination in TCHARs.
///
/// \return Same as wstrDestination .
////////////////////////////////////////////////////////////////////////////////
WCHAR* ANSIStringToWIDE( WCHAR *wstrDestination, const CHAR *strSource, 
						 int iDestChar )
{
	if ( wstrDestination == NULL || strSource == NULL )
		return NULL;

	if( iDestChar == -1 )
		iDestChar = (int) strlen( strSource );;

	MultiByteToWideChar( CP_ACP, 0, strSource, -1, 
		wstrDestination, iDestChar );

	wstrDestination[iDestChar] = 0;

	return wstrDestination;
} // ANSIStringToWide

////////////////////////////////////////////////////////////////////////////////
/// \brief Converts ANSI string to TCHAR.
///
/// UNICODE conversion utility to convert a ANSI string to a TCHAR string.
///
/// \param tstrDestination Resulting TCHAR string 
///        (not allocated by this function).
/// \param strSource Source CHAR string.
/// \param iDestChar Size of tstrDestination in TCHARs.
///
/// \return Same as tstrDestination .
////////////////////////////////////////////////////////////////////////////////
TCHAR* ANSIStringToGeneric( TCHAR *tstrDestination, const CHAR *strSource, 
						    int iDestChar = -1 )
{
	if ( tstrDestination == NULL || strSource == NULL || !iDestChar )
		return NULL;

#ifdef _UNICODE
	return ANSIStringToWIDE( tstrDestination, strSource, iDestChar );
#else
	if( iDestChar == -1 )
		strcpy( tstrDestination, strSource );	
	else
	{
		strncpy( tstrDestination, strSource, iDestChar );
		tstrDestination[iDestChar] = '\0';
	}

	return tstrDestination;
#endif    
} // ANSIStringToGeneric

#endif

////////////////////////////////////////////////////////////////////////////////
/// \brief Class constructor.
///
/// Initialize member variables to default values.
////////////////////////////////////////////////////////////////////////////////
CIntManager::CIntManager( )
{
	m_iLeft = m_iTop = m_iRight = m_iBottom = 0;
	m_stlObjList.clear( );

	m_pBindings        = NULL;
	m_szVSDeform       = NULL;
	m_szPSDeform       = NULL;

#if defined(OGL2)
	m_szInputVSDeform  = NULL;
	m_szInputPSDeform  = NULL;
	m_szReturnPSDeform = NULL;
#endif

	m_iNumMRTs         = 0;
	m_iNumRTs          = 0;
	m_iNumRPs          = 0;

	m_fZNear           = 0.1f;
	m_fZFar            = 1000.0f;

	m_bRenderIdx       = false;
	m_bAttDepth        = false;
	m_bAttTexCoord     = false;
	m_bAttTBN          = false;
	m_bAttUserDefI     = false;
	m_bAttUserDefNI    = false;
	m_bAttNormal       = false;
	m_bAttCurv         = false;
	m_bAttTensor2      = false;
	m_bAttTensor3      = false;
	m_iMaxVtCacheTex   = 0;

#if defined(D3D9)
	m_pd3dDevice = NULL;

	m_bInitDOCalled       = false;
	m_bRestoreDOCalled    = false;
	m_bInvalidateDOCalled = false;
	m_bDeleteDOCalled     = false;

	for( int iCt = 0; iCt < 16; iCt++ )
	{
		m_pRTT[iCt] = NULL;
		m_pRTS[iCt] = NULL;
	}

	m_pReadSurf        = NULL;
	m_pd3dBackBuffer   = NULL;

	// tbn.fx
	m_pTBNFx           = NULL;
	m_hTBNAdjFaceIdxC  = NULL;
	m_hTBNVtIdxC       = NULL;
    m_hTBNFaceIdxC     = NULL;

	// tensor.fx
	m_pTensorFx        = NULL;
	m_hTensorVtIdxC    = NULL;
	m_hTensorAdjVtIdxC = NULL;

	// n.fx
	m_pNFx             = NULL;
	m_hNAdjFaceIdxC    = NULL;
	m_hNVtIdxC         = NULL;
    m_hNFaceIdxC       = NULL;

#elif defined(OGL2)
	m_uiFBO = 0;

	// Initialize GLEW
	// It is expected that a valid OpenGL rendering context was created
	// beforehand
	GLenum eErr = glewInit( );
	if( GLEW_OK != eErr )
	{
		// Problem: glewInit failed, something is seriously wrong.
		const GLubyte *teste = glewGetErrorString( eErr );
		fprintf( stderr, "Error: %s\n", glewGetErrorString( eErr ) );
		m_bIsOK = false;
		return;
	}

    // Ensure we have the necessary OpenGL Shading Language extensions
    if( glewGetExtension("GL_ARB_fragment_shader")      != GL_TRUE ||
        glewGetExtension("GL_ARB_vertex_shader")        != GL_TRUE ||
        glewGetExtension("GL_ARB_shader_objects")       != GL_TRUE ||
        glewGetExtension("GL_ARB_shading_language_100") != GL_TRUE )
    {
        fprintf( stderr, "Driver does not support OpenGL Shading Language\n" );
		m_bIsOK = false;
		return;
    }

	for( int iCt = 0; iCt < 16; iCt++ )
	{
		m_eRTs[iCt]  = GL_COLOR_ATTACHMENT0_EXT + iCt;
		m_uiRTs[iCt] = 0;
	}

	// facetbn.glsl
	m_hFaceTBNPO               = 0;
	m_iFaceTBNVtMapSize        = 0;

	// vertextbn.glsl
	m_hVertexTBNPO             = 0;
	m_iVertexTBNAdjFaceMapSize = 0;
	m_iVertexTBNFaceMapSize    = 0;

	// tensor2.glsl
	m_h2ndTensorPO             = 0;
	m_i2ndTensorVtMapSize      = 0;
	m_i2ndTensorAdjVtMapSize   = 0;

	// tensor3.glsl
	m_h3rdTensorPO             = 0;
	m_i3rdTensorVtMapSize      = 0;
	m_i3rdTensorAdjVtMapSize   = 0;

	// facen.glsl
	m_hFaceNPO                 = 0;
	m_iFaceNVtMapSize          = 0;

	// vertexn.glsl
	m_hVertexNPO               = 0;
	m_iVertexNAdjFaceMapSize   = 0;
	m_iVertexNFaceMapSize      = 0;

	if( CreateRTs( ) )
	{
		m_bIsOK = false;
		return;
	}
#endif

	m_bIsOK = true;
} // CIntManager::CIntManager

////////////////////////////////////////////////////////////////////////////////
/// \brief Class destructor.
///
/// Delete resources allocated in the constructor.
////////////////////////////////////////////////////////////////////////////////
CIntManager::~CIntManager( )
{
	if( !m_bIsOK ) return;

	// Release each remaining object
	while( !m_stlObjList.empty( ) )
	{
		stlObjListIt stlIt = m_stlObjList.begin( );
		DestroyObject( (*stlIt) );
	}

	SAFE_DELETE_ARRAY( m_pBindings  );
	SAFE_DELETE_ARRAY( m_szVSDeform );
	SAFE_DELETE_ARRAY( m_szPSDeform );

#if defined(OGL2)
	SAFE_DELETE_ARRAY( m_szInputVSDeform  );
	SAFE_DELETE_ARRAY( m_szInputPSDeform  );
	SAFE_DELETE_ARRAY( m_szReturnPSDeform );
#endif
} // CIntManager::~CIntManager

////////////////////////////////////////////////////////////////////////////////
/// \brief Bind vertex semantics of the original geometry to 
///        interaction-specific vertex semantics.
///
/// \param pAttributes An array of ATTRIBUTEs, describing which attributes
///        should be computed per-pixel by the architecture of interaction. 
///        The array must be ended by the ATTTYPE_END definition.
///
/// \return 0 on success.
///         -1 on error (not enough memory).
////////////////////////////////////////////////////////////////////////////////
int CIntManager::SetAttributes( LPATTRIBUTE pAttributes )
{
	m_bAttDepth     = false;
	m_bAttTexCoord  = false;
	m_bAttTBN       = false;
	m_bAttUserDefI  = false;
	m_bAttUserDefNI = false;
	m_bAttNormal    = false;
	m_bAttCurv      = false;
	m_bAttTensor2   = false;
	m_bAttTensor3   = false;

	for( int iCt = 0; iCt < ATTTYPE_SIZE; iCt++ )
	{
		m_pAttEncoding[iCt].m_bActive = false;
		m_pAttEncoding[iCt].m_eType = iCt;
	}

	// View which attributes were selected
	int iNumAttributes = 0;
	while( pAttributes[iNumAttributes].iAttribute != ATTTYPE_END )
	{
		switch( pAttributes[iNumAttributes].iAttribute )
		{
		case ATTTYPE_DEPTH    : m_bAttDepth     = true; break;
		case ATTTYPE_TEXCOORD : m_bAttTexCoord  = true; break;
		case ATTTYPE_TBN      : m_bAttTBN       = true; break;
		case ATTTYPE_USERDEFI : m_bAttUserDefI  = true; break;
		case ATTTYPE_USERDEFNI: m_bAttUserDefNI = true; break;
		case ATTTYPE_NORMAL   : m_bAttNormal    = true; break;
		case ATTTYPE_CURV     : m_bAttCurv      = true; break;
		case ATTTYPE_TENSOR2  : m_bAttTensor2   = true; break;
		case ATTTYPE_TENSOR3  : m_bAttTensor3   = true; break;
		}
		iNumAttributes++;
	}

	LPATTENCODING pAtt; // Pointer to the current attribute
	int iCurAtt = 0;    // Current attribute
	int iRP     = 0;    // Render pass
	int iComp   = 0;    // Attribute component between 0 and 3
	int iRT     = 0;    // Render target between 0 and m_iNumRTs-1

	// Like iRT, but is zeroed for each new render pass
	int iCurRT = 0;

	// Number of texture samplers in the vertex shader
	int iVtSamplers = 1;

	// Depth
	if( m_bAttDepth )
	{
		iCurAtt = ATTTYPE_DEPTH;

		// Components must fit in the same render pass
		// depth (1)
		if( 1 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+0 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0;
			iVtSamplers = 0;
		}

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iNumAtts = 1;
		pAtt->m_iRP = iRP;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}

	// Texture coordinates
	if( m_bAttTexCoord )
	{
		iCurAtt = ATTTYPE_TEXCOORD;

		// Components must fit in the same render pass
		// tex coords (2)
		if( 2 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+0 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0; 
			iVtSamplers = 0;
		}

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iNumAtts = 2;
		pAtt->m_iRP = iRP;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[1] = iComp; pAtt->m_iCompRT[1] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}

	// TBN frame
	if( m_bAttTBN )
	{
		iCurAtt = ATTTYPE_TBN;

		// Components must fit in the same render pass
		// tangent (3), bitangent (3)
		if( 6 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+1 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0;
			iVtSamplers = 0;
		}

		iVtSamplers++;

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iRP = iRP;
		pAtt->m_iNumAtts = 6;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[1] = iComp; pAtt->m_iCompRT[1] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[2] = iComp; pAtt->m_iCompRT[2] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[3] = iComp; pAtt->m_iCompRT[3] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[4] = iComp; pAtt->m_iCompRT[4] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[5] = iComp; pAtt->m_iCompRT[5] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}

	// User-defined value for indexed geometry
	if( m_bAttUserDefI )
	{
		iCurAtt = ATTTYPE_USERDEFI;

		// Components must fit in the same render pass
		// (4)
		if( 4 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+0 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0;
			iVtSamplers = 0;
		}

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iNumAtts = 4;
		pAtt->m_iRP = iRP;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[1] = iComp; pAtt->m_iCompRT[1] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[2] = iComp; pAtt->m_iCompRT[2] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[3] = iComp; pAtt->m_iCompRT[3] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}

	// Normal vector
	if( m_bAttNormal )
	{
		iCurAtt = ATTTYPE_NORMAL;

		// Components must fit in the same render pass
		if( 3 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+1 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0;
			iVtSamplers = 0;
		}

		iVtSamplers++;

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iNumAtts = 3;
		pAtt->m_iRP = iRP;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[1] = iComp; pAtt->m_iCompRT[1] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[2] = iComp; pAtt->m_iCompRT[2] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}
	
	// Curvature
	if( m_bAttCurv )
	{
		iCurAtt = ATTTYPE_CURV;

		// Components must fit in the same render pass
		// k1 (1), k1 (1), pdir1 (3), pdir2 (3)
		if( 8 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+2 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0;
			iVtSamplers = 0;
		}

		iVtSamplers += 2;

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iNumAtts = 8;
		pAtt->m_iRP = iRP;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[1] = iComp; pAtt->m_iCompRT[1] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[2] = iComp; pAtt->m_iCompRT[2] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[3] = iComp; pAtt->m_iCompRT[3] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[4] = iComp; pAtt->m_iCompRT[4] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[5] = iComp; pAtt->m_iCompRT[5] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[6] = iComp; pAtt->m_iCompRT[6] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[7] = iComp; pAtt->m_iCompRT[7] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}

	// 2nd fundamental form
	if( m_bAttTensor2 )
	{
		iCurAtt = ATTTYPE_TENSOR2;

		// Components must fit in the same render pass
		// 2nd fundamental form e,f,g (3)
		if( 3 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+1 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0;
			iVtSamplers = 0;
		}

		iVtSamplers++;

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iNumAtts = 3;
		pAtt->m_iRP = iRP;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[1] = iComp; pAtt->m_iCompRT[1] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[2] = iComp; pAtt->m_iCompRT[2] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}

	// Tensor of curvature derivative
	if( m_bAttTensor3 )
	{
		iCurAtt = ATTTYPE_TENSOR3;

		// Components must fit in the same render pass
		// Tensor of curvature derivative form a,b,c,d (4)
		if( 4 > ( m_iNumMRTs - ( iRT % m_iNumMRTs ) ) * 4 - iComp ||
			iVtSamplers+1 > 4 )
		{ 
			m_pRenderPass[iRP].m_iRP    = iRP;
			m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
			m_pRenderPass[iRP].m_iComps = 4 - iComp;
			iComp = 0; iRT++; iRP++; iCurRT = 0;
			iVtSamplers = 0;
		}

		iVtSamplers++;

		pAtt = &m_pAttEncoding[iCurAtt];
		pAtt->m_iNumAtts = 4;
		pAtt->m_iRP = iRP;
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[0] = iComp; pAtt->m_iCompRT[0] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[1] = iComp; pAtt->m_iCompRT[1] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[2] = iComp; pAtt->m_iCompRT[2] = iRT;
		iComp++; 
		if( iComp == 4 ) { iComp = 0; iRT++; iCurRT++; };
		pAtt->m_iComp[3] = iComp; pAtt->m_iCompRT[3] = iRT;
		iComp++; 

		pAtt->m_bActive = true;
	}

	m_pRenderPass[iRP].m_iRP    = iRP;
	m_pRenderPass[iRP].m_iRTs   = iCurRT + 1;
	m_pRenderPass[iRP].m_iComps = 4 - iComp;

	if( m_bAttDepth    || m_bAttTexCoord || m_bAttTBN  ||
		m_bAttUserDefI || m_bAttNormal   || m_bAttCurv || 
		m_bAttTensor2  || m_bAttTensor3 )
		m_bRenderIdx = true;
	else
		m_bRenderIdx = false;

	// Attributes obtained with special render modes
	if( m_bAttUserDefNI && m_bRenderIdx )
		iRT++;
	
	// Re-initialize the shaders
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		 stlIt != m_stlObjList.end( ); stlIt++ )
	{
		if( (*stlIt)->DestroyShaders( ) )
			return -1;

		if( (*stlIt)->CreateShaders( ) )
			return -1;
	}

	m_iNumRTs = iRT + 1;
	m_iNumRPs = iRP + 1;

	// Determine number of cache textures for vertex-level computations
	m_iMaxVtCacheTex = 0;

	// Face normal (1 RT)
	if( m_bAttNormal ) m_iMaxVtCacheTex = 1;
	// Anything that needs normals
	if( m_bAttCurv || m_bAttTensor2 || m_bAttTensor3 )
		m_iMaxVtCacheTex = 1;
	// Face TBN (3 RTs)
	if( m_bAttTBN ) m_iMaxVtCacheTex = 3;

	// Re-initialize the render targets
	DestroyRTs( );
	CreateRTs ( );

	for( stlObjListIt stlIt = m_stlObjList.begin( );
		 stlIt != m_stlObjList.end( ); stlIt++ )
	{
		if( (*stlIt)->DestroyRTs( ) )
			return -1;

		if( (*stlIt)->CreateRTs( ) )
			return -1;
	}

	return 0;
} // CIntManager::SetAttributes

////////////////////////////////////////////////////////////////////////////////
/// \brief Bind vertex semantics of the original geometry to 
///        interaction-specific vertex semantics.
///
/// \param pBindings An array of SEMANTICBINDINGs, describing how each input
///        vertex semantic is assigned to an interaction-specific vertex
///        semantic. The array must be ended by the SEMANTIC_END definition.
///
/// \return 0 on success.
///         -1 on error.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::BindSemantics( LPSEMANTICBINDING pBindings )
{
	int iCt;

	// Get number of bindings
	int iNumBindings = 0;
	while( pBindings[iNumBindings++].iVSSemantic != VSSEMANTIC_END );

	iNumBindings--;

	if( iNumBindings < 1 )
		return -1;

	// Check for duplications of VSSEMANTIC values
	for( iCt = 0; iCt < iNumBindings; iCt++ )
	{
		VSSEMANTIC eCurValue = pBindings[iCt].iVSSemantic;

		for( int iCt2 = 0; iCt2 < iNumBindings; iCt2++ )
		{
			if( iCt2 != iCt )
			{
				if( eCurValue == pBindings[iCt2].iVSSemantic )
					return -1;
			}
		}
	}

	// Re-allocate memory for our array of semantic bindings
	SAFE_DELETE_ARRAY( m_pBindings );
	m_pBindings = new SEMANTICBINDING[ iNumBindings + 1 ];
	if( m_pBindings == NULL )
		return -1;

	// Make a copy of the bindings and check
	// whether they are valid
	for( iCt = 0; iCt < iNumBindings; iCt++ )
	{
		if( pBindings[iCt].iVSSemantic != VSSEMANTIC_END )
		{
			if( pBindings[iCt].iVSSemantic < 0 ||
				pBindings[iCt].iVSSemantic >= VSSEMANTIC_MAXVALUE )
				return -1;

			if( pBindings[iCt].iIntSemantic < 0 ||
				pBindings[iCt].iIntSemantic >= INTSEMANTIC_MAXVALUE )
				return -1;
		}

		m_pBindings[iCt] = pBindings[iCt];
	}

	m_pBindings[ iNumBindings ].iVSSemantic = VSSEMANTIC_END;

	// Create text string with the definition of the VSDEFORM 
	// data structure
	SAFE_DELETE_ARRAY( m_szVSDeform );
	m_szVSDeform = new char[32 + ( iNumBindings + 1 ) * 32];
	if( m_szVSDeform == NULL )
		return -1;
	m_szVSDeform[0] = 0x00;

	// Create text string with the definition of the PSDEFORM
	// data structure
	SAFE_DELETE_ARRAY( m_szPSDeform );
	m_szPSDeform = new char[32 + ( iNumBindings + 18 + 1 ) * 32];
	if( m_szPSDeform == NULL )
		return -1;
	m_szPSDeform[0] = 0x00;

#if defined(OGL2)
	SAFE_DELETE_ARRAY( m_szInputVSDeform );
	m_szInputVSDeform = new char[32 + ( iNumBindings + 16 + 1 ) * 32];
	if( m_szInputVSDeform == NULL )
		return -1;
	m_szInputVSDeform[0] = 0x00;

	SAFE_DELETE_ARRAY( m_szInputPSDeform );
	m_szInputPSDeform = new char[32 + ( iNumBindings + 16 + 1 ) * 32];
	if( m_szInputPSDeform == NULL )
		return -1;
	m_szInputPSDeform[0] = 0x00;

	SAFE_DELETE_ARRAY( m_szReturnPSDeform );
	m_szReturnPSDeform = new char[32 + ( iNumBindings + 16 + 1 ) * 32];
	if( m_szReturnPSDeform == NULL )
		return -1;
	m_szReturnPSDeform[0] = 0x00;

	strcat( m_szPSDeform, "varying vec4 g_vNormal;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vPosition;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vTexCoord8;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vTexCoord9;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vTexCoord10;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vTexCoord11;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vTexCoord12;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vTexCoord13;\n" );
	strcat( m_szPSDeform, "varying vec4 g_vTexCoord14;\n" );
#endif

	strcat( m_szVSDeform, "struct VSDEFORM\n{\n" );
	strcat( m_szPSDeform, "struct PSDEFORM\n{\n" );

#if defined(D3D9)
	strcat( m_szVSDeform, "\tfloat4 vPos    : POSITION;\n" );
	strcat( m_szPSDeform, "\tfloat4 vPos    : POSITION;\n" );
#elif defined(OGL2)
	strcat( m_szVSDeform, "\tvec4 vPos;\n" );
	strcat( m_szPSDeform, "\tvec4 vPos;\n" );
	strcat( m_szInputVSDeform, "\tIn.vPos = gl_Vertex;\n" );
	strcat( m_szInputPSDeform, "\tDefOut.vPos = g_vPosition;\n" );
	strcat( m_szReturnPSDeform, "\tgl_Position = Out.vPos;\n" );
	strcat( m_szReturnPSDeform, "\tg_vPosition = Out.vPos;\n" );
	strcat( m_szReturnPSDeform, "\tg_vNormal = vec4(0);\n" );
#endif

	for( iCt = 0; iCt < iNumBindings; iCt++ )
	{
#if defined(D3D9)
		switch( m_pBindings[iCt].iVSSemantic )
		{
		case VSSEMANTIC_NORMAL   : 
			strcat( m_szVSDeform, "\tfloat4 vNormal : NORMAL;\n"    );
			strcat( m_szPSDeform, "\tfloat4 vNormal : NORMAL;\n"    );
			break;
		case VSSEMANTIC_COLOR0   : 
			strcat( m_szVSDeform, "\tfloat4 vColor0 : COLOR0;\n"    );
			strcat( m_szPSDeform, "\tfloat4 vColor0 : COLOR0;\n"    );
			break;
		case VSSEMANTIC_COLOR1   : 
			strcat( m_szVSDeform, "\tfloat4 vColor1 : COLOR1;\n"    );
			strcat( m_szPSDeform, "\tfloat4 vColor1 : COLOR1;\n"    );
			break;
		case VSSEMANTIC_TEXCOORD0: 
			strcat( m_szVSDeform, "\tfloat4 vTex0   : TEXCOORD0;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex0   : TEXCOORD0;\n" );
			break;
		case VSSEMANTIC_TEXCOORD1: 
			strcat( m_szVSDeform, "\tfloat4 vTex1   : TEXCOORD1;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex1   : TEXCOORD1;\n" );
			break;
		case VSSEMANTIC_TEXCOORD2: 
			strcat( m_szVSDeform, "\tfloat4 vTex2   : TEXCOORD2;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex2   : TEXCOORD2;\n" );
			break;
		case VSSEMANTIC_TEXCOORD3: 
			strcat( m_szVSDeform, "\tfloat4 vTex3   : TEXCOORD3;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex3   : TEXCOORD3;\n" );
			break;
		case VSSEMANTIC_TEXCOORD4: 
			strcat( m_szVSDeform, "\tfloat4 vTex4   : TEXCOORD4;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex4   : TEXCOORD4;\n" );
			break;
		case VSSEMANTIC_TEXCOORD5: 
			strcat( m_szVSDeform, "\tfloat4 vTex5   : TEXCOORD5;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex5   : TEXCOORD5;\n" );
			break;
		case VSSEMANTIC_TEXCOORD6: 
			strcat( m_szVSDeform, "\tfloat4 vTex6   : TEXCOORD6;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex6   : TEXCOORD6;\n" );
			break;
		case VSSEMANTIC_TEXCOORD7: 
			strcat( m_szVSDeform, "\tfloat4 vTex7   : TEXCOORD7;\n" );
			strcat( m_szPSDeform, "\tfloat4 vTex7   : TEXCOORD7;\n" );
			break;
		}
#elif defined(OGL2)
		switch( m_pBindings[iCt].iVSSemantic )
		{
		case VSSEMANTIC_NORMAL   : 
			strcat( m_szVSDeform, "\tvec4 vNormal;\n"  );
			strcat( m_szPSDeform, "\tvec4 vNormal;\n"  );
			strcat( m_szInputVSDeform, "\tIn.vNormal = gl_Normal;\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vNormal = g_vNormal;\n" );
			strcat( m_szReturnPSDeform, "\tg_vNormal = Out.vNormal;\n" );
			break;
		case VSSEMANTIC_COLOR0   : 
			strcat( m_szVSDeform, "\tvec4 vColor0;\n"  );
			strcat( m_szPSDeform, "\tvec4 vColor0;\n"  );
			strcat( m_szInputVSDeform, "\tIn.vColor0 = gl_Color;\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vColor0 = gl_Color;\n" );
			strcat( m_szReturnPSDeform, "\tgl_FrontColor = Out.vColor0;\n" );
			break;
		case VSSEMANTIC_COLOR1   : 
			strcat( m_szVSDeform, "\tvec4 vColor1;\n"  );
			strcat( m_szPSDeform, "\tvec4 vColor1;\n"  );
			strcat( m_szInputVSDeform, "\tIn.vColor1 = gl_SecondaryColor;\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vColor1 = gl_SecondaryColor;\n" );
			strcat( m_szReturnPSDeform, "\tgl_BackColor = Out.vColor1;\n" );
			break;
		case VSSEMANTIC_TEXCOORD0: 
			strcat( m_szVSDeform, "\tvec4 vTex0;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex0;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex0 = gl_TexCoord[0];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex0 = gl_TexCoord[0];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[0] = Out.vTex0;\n" );
			break;
		case VSSEMANTIC_TEXCOORD1: 
			strcat( m_szVSDeform, "\tvec4 vTex1;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex1;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex1 = gl_TexCoord[1];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex1 = gl_TexCoord[1];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[1] = Out.vTex1;\n" );
			break;
		case VSSEMANTIC_TEXCOORD2: 
			strcat( m_szVSDeform, "\tvec4 vTex2;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex2;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex2 = gl_TexCoord[2];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex2 = gl_TexCoord[2];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[2] = Out.vTex2;\n" );
			break;
		case VSSEMANTIC_TEXCOORD3: 
			strcat( m_szVSDeform, "\tvec4 vTex3;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex3;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex3 = gl_TexCoord[3];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex3 = gl_TexCoord[3];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[3] = Out.vTex3;\n" );
			break;
		case VSSEMANTIC_TEXCOORD4: 
			strcat( m_szVSDeform, "\tvec4 vTex4;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex4;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex4 = gl_TexCoord[4];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex4 = gl_TexCoord[4];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[4] = Out.vTex4;\n" );
			break;
		case VSSEMANTIC_TEXCOORD5: 
			strcat( m_szVSDeform, "\tvec4 vTex5;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex5;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex5 = gl_TexCoord[5];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex5 = gl_TexCoord[5];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[5] = Out.vTex5;\n" );
			break;
		case VSSEMANTIC_TEXCOORD6: 
			strcat( m_szVSDeform, "\tvec4 vTex6;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex6;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex6 = gl_TexCoord[6];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex6 = gl_TexCoord[6];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[6] = Out.vTex6;\n" );
			break;
		case VSSEMANTIC_TEXCOORD7: 
			strcat( m_szVSDeform, "\tvec4 vTex7;\n"    );
			strcat( m_szPSDeform, "\tvec4 vTex7;\n"    );
			strcat( m_szInputVSDeform, "\tIn.vTex7 = gl_TexCoord[7];\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTex7 = gl_TexCoord[7];\n" );
			strcat( m_szReturnPSDeform, "\tgl_TexCoord[7] = Out.vTex7;\n" );
			break;
		}
#endif
	}
	strcat( m_szVSDeform, "};\n" );

	// Finish adding attributes to PSDEFORM
#if defined(D3D9)
	strcat( m_szPSDeform, "\tfloat4 vTPos : TEXCOORD8;\n" );

	if( m_bAttTBN || m_bAttNormal )
	{
		strcat( m_szPSDeform, "\tfloat3 vTN : TEXCOORD9;\n" );

		if( m_bAttTBN )
		{
			strcat( m_szPSDeform, "\tfloat3 vTT : TEXCOORD10;\n" );
			strcat( m_szPSDeform, "\tfloat3 vTB : TEXCOORD11;\n" );
		}
	}

	if( m_bAttCurv )
	{
		strcat( m_szPSDeform, "\tfloat4 vP1K1 : TEXCOORD12;\n" );
		strcat( m_szPSDeform, "\tfloat4 vP2K2 : TEXCOORD13;\n" );
	}

	if( m_bAttTensor2 )
		strcat( m_szPSDeform, "\tfloat3 vTensor2 : TEXCOORD14;\n" );

	if( m_bAttTensor3 )
		strcat( m_szPSDeform, "\tfloat4 vTensor3 : TEXCOORD15;\n" );

#elif defined(OGL2)
	strcat( m_szPSDeform, "\tvec4 vTPos;\n" );
	strcat( m_szInputPSDeform, "\tDefOut.vTPos = g_vTexCoord8;\n" );
	strcat( m_szReturnPSDeform, "\tg_vTexCoord8 = Out.vTPos;\n" );

	if( m_bAttTBN || m_bAttNormal )
	{
		strcat( m_szPSDeform, "\tvec3 vTN;\n" );
		strcat( m_szInputPSDeform, "\tDefOut.vTN = g_vTexCoord9.xyz;\n" );
		strcat( m_szReturnPSDeform, "\tg_vTexCoord9 = Out.vTN;\n" );

		if( m_bAttTBN )
		{
 			strcat( m_szPSDeform, "\tvec3 vTT;\n" );
			strcat( m_szPSDeform, "\tvec3 vTB;\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTT = g_vTexCoord10.xyz;\n" );
			strcat( m_szReturnPSDeform, "\tg_vTexCoord10 = Out.vTT;\n" );
			strcat( m_szInputPSDeform, "\tDefOut.vTB = g_vTexCoord11.xyz;\n" );
			strcat( m_szReturnPSDeform, "\tg_vTexCoord11 = Out.vTB;\n" );
		}
	}

	if( m_bAttCurv )
	{
		strcat( m_szPSDeform, "\tvec4 vP1K1;\n" );
		strcat( m_szPSDeform, "\tvec4 vP2K2;\n" );
		strcat( m_szInputPSDeform, "\tDefOut.vP1K1 = g_vTexCoord12;\n" );
		strcat( m_szReturnPSDeform, "\tg_vTexCoord12 = Out.vP1K1;\n" );
		strcat( m_szInputPSDeform, "\tDefOut.vP2K2 = g_vTexCoord13;\n" );
		strcat( m_szReturnPSDeform, "\tg_vTexCoord13 = Out.vP2K2;\n" );
	}

	if( m_bAttTensor2 )
	{
		strcat( m_szPSDeform, "\tvec3 vTensor2;\n" );
		strcat( m_szInputPSDeform, "\tDefOut.vTensor2 = g_vTexCoord14.xyz;\n" );
		strcat( m_szReturnPSDeform, "\tg_vTexCoord14 = Out.vTensor2;\n" );
	}

	if( m_bAttTensor3 )
	{
		strcat( m_szPSDeform, "\tvec4 vTensor3;\n" );
		strcat( m_szInputPSDeform, "\tDefOut.vTensor3 = g_vTexCoord15;\n" );
		strcat( m_szReturnPSDeform, "\tg_vTexCoord15 = Out.vTensor3;\n" );
	}
#endif

	strcat( m_szPSDeform, "};\n" );

	// Re-initialize the shaders
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		 stlIt != m_stlObjList.end( ); stlIt++ )
	{
		if( (*stlIt)->DestroyShaders( ) )
			return -1;

		if( (*stlIt)->CreateShaders( ) )
			return -1;
	}

	return 0;
} // CIntManager::BindSemantics

////////////////////////////////////////////////////////////////////////////////
/// \brief Set the region of interest for obtaining the geometry attributes 
///        computed by the architecture of interaction.
///
/// The ROI is a rectangle given in coordinates relative to the viewport, 
/// in pixels. Values are clamped to the viewport boundaries.
///
/// \param iLeft X-coordinate of the upper-left corner of the ROI.
/// \param iTop Y-coordinate of the upper-left corner of the ROI.
/// \param iRight X-coordinate of the lower-right corner of the ROI.
/// \param iBottom Y-coordinate of the lower-right corner of the ROI.
///
/// \return  0 on success.
///         -1 if the D3D device is invalid (D3D only).
///         -2 if an error occurred when invalidating or restore the
///            device objects (D3D only).
///         -3 if CIntManager::m_bIsOK is false.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::SetROI( int iLeft, int iTop, int iRight, int iBottom )
{
	if( !m_bIsOK ) return -3;

	int iMinX = 0, iMaxX = 0, iMinY = 0, iMaxY = 0;

#if defined(D3D9)
	if( m_pd3dDevice == NULL )
		return -1;

	// Get viewport rectangle for Direct3D
	D3DVIEWPORT9 d3dViewport;
	m_pd3dDevice->GetViewport( &d3dViewport );

	iMinX = d3dViewport.X;
	iMinY = d3dViewport.Y;
	iMaxX = iMinX + d3dViewport.Width  - 1;
	iMaxY = iMinY + d3dViewport.Height - 1;
#elif defined(OGL2)
	// Get viewport rectangle for OpenGL
	GLint vViewport[4];
	glGetIntegerv( GL_VIEWPORT, vViewport );

	iMinX = vViewport[0];
	iMinY = vViewport[1];
	iMaxX = iMinX + vViewport[2] - 1;
	iMaxY = iMinY + vViewport[3] - 1;
#endif

	// Save last ROI's width and height
	int iLastROIWidth  = m_iRight  - m_iLeft + 1;
	int iLastROIHeight = m_iBottom - m_iTop  + 1;

	// Set new ROI dimensions
	m_iLeft   = SATURATE_MIN( iMinX, SATURATE_MAX( iMaxX, iLeft   ) );
	m_iTop    = SATURATE_MIN( iMinY, SATURATE_MAX( iMaxY, iTop    ) );
	m_iRight  = SATURATE_MIN( iMinX, SATURATE_MAX( iMaxX, iRight  ) );
	m_iBottom = SATURATE_MIN( iMinY, SATURATE_MAX( iMaxY, iBottom ) );

	if( m_iRight  < m_iLeft ) m_iRight  = m_iLeft;
	if( m_iBottom < m_iTop  ) m_iBottom = m_iTop;

	// Has the width or height changed?
	int iCurROIWidth  = m_iRight  - m_iLeft + 1;
	int iCurROIHeight = m_iBottom - m_iTop  + 1;

	if( iLastROIWidth != iCurROIWidth || iLastROIHeight != iCurROIHeight )
	{
		// Need to re-create render target textures
		if( DestroyRTs( ) )	return -2;
		if( CreateRTs ( ) ) return -2;
	}

	// Compute new projection matrix
	ComputeROIMatrix( );

	return 0;
} // CIntManager::SetROI

////////////////////////////////////////////////////////////////////////////////
/// \brief Retrieve the coordinates of the current region of interest (ROI).
///
/// \param iLeft Reference to x-coord. of the upper-left corner of the ROI.
/// \param iTop Reference to y-coord of the upper-left corner of the ROI.
/// \param iRight Reference to x-coord. of the lower-right corner of the ROI.
/// \param iBottom Reference to y-coord. of the lower-right corner of the ROI.
///
/// \return None.
////////////////////////////////////////////////////////////////////////////////
void CIntManager::GetROI( int &iLeft, int &iTop, int &iRight, int &iBottom )
{
	if( !m_bIsOK ) return;

	iLeft   = m_iLeft;
	iTop    = m_iTop;
	iRight  = m_iRight;
	iBottom = m_iBottom;
} // CIntManager::GetROI

////////////////////////////////////////////////////////////////////////////////
/// \brief Set the z-value of the near clipping plane, used by the architecture
///        of interaction when rendering the attributes in the region of 
///        interest.
///
/// \param fZNear Z-value of the near clipping plane.
///
/// \return None.
////////////////////////////////////////////////////////////////////////////////
void CIntManager::SetZNear( float fZNear )
{
	if( fZNear != m_fZNear )
	{
		m_fZNear = fZNear;
		ComputeROIMatrix( );
	}
} // CIntManager::SetZNear

////////////////////////////////////////////////////////////////////////////////
/// \brief Set the z-value of the far clipping plane, used by the architecture
///        of interaction when rendering the attributes in the region of 
///        interest.
///
/// \param fZFar Z-value of the far clipping plane.
///
/// \return None.
////////////////////////////////////////////////////////////////////////////////
void CIntManager::SetZFar( float fZFar )
{
	if( fZFar != m_fZFar )
	{
		m_fZFar = fZFar;
		ComputeROIMatrix( );
	}
} // CIntManager::SetZFar

////////////////////////////////////////////////////////////////////////////////
/// \brief Create a new object for interaction.
///
/// This creates an entity related to a set of graphics primitives defined by
/// the user (a geometric model), and contains the data required for performing 
/// the processing by the architecture of interaction.
///
/// \param pObj Pointer to pointer of CIntObj that will receive the new object 
///        created.
/// \param szAdjWriteFile Name of the file where the adjacency data is to
///        be written. This parameter is optional and should be set to NULL
///        when not used.
/// \param szAdjReadFile Name of the file from which the adjacency data is to
///        be read. This parameter is optional and should be set to NULL when
///        not used.
///
/// \return 0 on success.
///         -1 if the D3D device is invalid (D3D only).
///         -2 if there is not enough memory to create CIntObj.
///         -3 if there was an initialization error.
///         -4 if CIntManager::m_bIsOK is false.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::CreateObject( CIntObj **pObj, const char *szAdjWriteFile, 
							   const char *szAdjReadFile )
{
	if( !m_bIsOK ) return -4;

#if defined(D3D9)
	if( m_pd3dDevice == NULL )
		return -1;
#endif

	// Create CIntObj object
	*pObj = new CIntObj( this, szAdjWriteFile, szAdjReadFile );
	if( *pObj == NULL )
		return -2;

	// Initialize resources of the created object 
#if defined(D3D9)
	if( FAILED( (*pObj)->InitDeviceObjects( ) ) )
	{
		SAFE_DELETE( *pObj );
		return -3;
	}
	if( FAILED( (*pObj)->RestoreDeviceObjects( ) ) )
	{
		SAFE_DELETE( (*pObj) );
		return -3;
	}
#elif defined(OGL2)
	if( (*pObj)->CreateRTs( ) )
		return -3;
#endif

	// Add to list of created objects
	m_stlObjList.push_back( (*pObj) );

	return 0;
} // CIntManager::CreateObject

////////////////////////////////////////////////////////////////////////////////
/// \brief Destroy an object created with CIntManager::CreateObject.
///
/// \param pObj Pointer to the CIntObj object to be destroyed.
///
/// \return 0 on success.
///         -1 if the D3D device is invalid (D3D only).
///         -2 if there was a finalization error.
///         -3 if the given object was not found in the list of created objects.
///         -4 if CIntManager::m_bIsOK is false.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::DestroyObject( CIntObj *pObj )
{
	if( !m_bIsOK ) return -4;

#if defined(D3D9)
	if( m_pd3dDevice == NULL )
		return -1;

	if( FAILED( pObj->InvalidateDeviceObjects( ) ) )
		return -2;
	if( FAILED( pObj->DeleteDeviceObjects( ) ) )
		return -2;
#elif defined(OGL2)
	if( pObj->DestroyRTs( ) )
		return -2;
#endif

	// Look for object in the list
	stlObjListIt stlIt = m_stlObjList.begin( );
	while( stlIt != m_stlObjList.end( ) )
	{
		if( pObj == (*stlIt) )
		{
			m_stlObjList.erase( stlIt ); // Remove from list
			SAFE_DELETE( pObj );         // Release memory

			return 0;
		}
		stlIt++;
	}

	return -3;
} // CIntManager::DestroyObject

////////////////////////////////////////////////////////////////////////////////
/// \brief Retrieve the number of created objects.
///
/// \return Number of current created objects.
///         -1 on error.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::GetNumObjects( ) 
{
	if( !m_bIsOK ) return -1;

	return (int) m_stlObjList.size( ); 
} // CIntManager::GetNumObjects

////////////////////////////////////////////////////////////////////////////////
/// \brief Retrieve the list of created objects.
///
/// \param pObjList Pointer to an array of CIntObj objects that will be filled
///        by the pointers of the current created objects. This array should
///        be at least of the size of CIntManager::GetNumObjects.
///
/// \return None.
////////////////////////////////////////////////////////////////////////////////
void CIntManager::GetObjects( CIntObj **pObjList )
{
	if( !m_bIsOK ) return;

	int iCt = 0;

	// Iterate the list of created objects
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		 stlIt != m_stlObjList.end( ); stlIt++, iCt++ )
	{
		pObjList[iCt] = (*stlIt);
	}
} // CIntManager::GetObjects

////////////////////////////////////////////////////////////////////////////////
/// \brief Render all objects to off-screen render targets, using a shader
///        that computes geometry attributes encoded in the pixels.
///
/// For D3D applications, this function should be called between the calls for
/// IDirect3DDevice9::BeginScene and IDirect3DDevice9::EndScene.
///
/// \return 0 on success.
///         -1 if the D3D device is invalid (D3D only).
///         -2 if a rendering error has occurred.
///         -3 if CIntManager::m_bIsOK is false.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::Render( )
{
	if( !m_bIsOK ) return -3;

#if defined(OGL2)
	// Save current states
	glPushAttrib( GL_ALL_ATTRIB_BITS );

	// Bind framebuffer object
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, m_uiFBO );
	glDrawBuffer( GL_COLOR_ATTACHMENT0_EXT );
	glClearColor( 0, 0, 0, 0 );
#endif

	// Clear each render target
	for( int iCt = 0; iCt < m_iNumRTs; iCt++ )
	{
#if defined(D3D9)
		m_pd3dDevice->SetRenderTarget( 0, m_pRTS[iCt] );
		m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET,
							 D3DCOLOR_ARGB( 0, 0, 0, 0 ), 1.0f, 0L );
#elif defined(OGL2)
		glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, 
								   GL_COLOR_ATTACHMENT0_EXT, 
								   GL_TEXTURE_2D, m_uiRTs[iCt], 0 );
    	glClear( GL_COLOR_BUFFER_BIT );
#endif
	}

	// Update objects' geometries and re-computes geometry attributes
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		 stlIt != m_stlObjList.end( ); stlIt++ )
		(*stlIt)->Update( );

	// DEBUG: Re-enable rendering to the back buffer
//	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

	// Render objects
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		 stlIt != m_stlObjList.end( ); stlIt++ )
		(*stlIt)->Render( );

#if defined(D3D9)
	// Disable shaders
	m_pd3dDevice->SetVertexShader( NULL );
	m_pd3dDevice->SetPixelShader ( NULL );

	// Re-enable rendering to the back buffer
	HRESULT hrResult = m_pd3dDevice->SetRenderTarget( 0, m_pd3dBackBuffer );
	if( FAILED( hrResult ) )
		return -2;

	// The remaining render targets are set to null
	for( int iCt = 1; iCt < 4; iCt++ )
	{
		hrResult = m_pd3dDevice->SetRenderTarget( iCt, NULL );
		if( FAILED( hrResult ) )
			return -2;
	}
#elif defined(OGL2)
	// Disable shaders
	glUseProgramObjectARB( 0 );
	
	// Re-enable rendering to the back buffer
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

	// Restore states
	glPopAttrib( );
#endif

	return 0;
} // CIntManager::Render

////////////////////////////////////////////////////////////////////////////////
/// \brief Render objects to the back buffer, for debugging purposes.
///
/// For D3D applications, this function should be called between the calls for
/// IDirect3DDevice9::BeginScene and IDirect3DDevice9::EndScene.
///
/// \param bOvlWire Whether to overlay a wireframe rendering over the
///        shaded one.
/// \param fAspect Aspect ratio of the projection.
/// \param fZNear Distance of the near clipping plane.
/// \param fZFar Distance of the far clipping plane.
///
/// \return 0 on success.
///         -1 if the D3D device is invalid (D3D only).
///         -2 if a rendering error has occurred.
///         -3 if CIntManager::m_bIsOK is false.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::RenderDebug( bool bOvlWire, float fAspect, float fZNear, 
							  float fZFar, 
#if defined(D3D9)
							  LPDIRECT3DTEXTURE9 pd3dActualTex,
							  LPDIRECT3DTEXTURE9 pd3dReliefTex,
							  LPDIRECT3DTEXTURE9 pd3dBumpTex,
#elif defined(OGL2)
							  GLuint uiActualTex, 
							  GLuint uiReliefTex, 
							  GLuint uidBumpTex,
#endif
							  bool bUseReliefMap, bool bUseBumpMap, 
							  int iShowAtt )
{
	if( !m_bIsOK ) return -3;

#if defined(D3D9)
	// Re-enable rendering to the back buffer
	HRESULT hrResult = m_pd3dDevice->SetRenderTarget( 0, m_pd3dBackBuffer );
	if( FAILED( hrResult ) )
		return -2;

	m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
						 D3DCOLOR_ARGB( 255, 0, 0, 0 ), 1.0f, 0L );
#elif defined(OGL2)
	// Re-enable rendering to the back buffer
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );

	glClearColor( 0, 0, 0, 1 );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
#endif

	// Render objects
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		stlIt != m_stlObjList.end( ); stlIt++ )
	{
		(*stlIt)->RenderDebug( bOvlWire, fAspect, fZNear, fZFar, 
#if defined(D3D9)
			pd3dActualTex, pd3dReliefTex, pd3dBumpTex,
#elif defined(OGL2)
			uiActualTex, uiReliefTex, uidBumpTex,
#endif
			bUseReliefMap, bUseBumpMap, iShowAtt );
	}

#if defined(D3D9)
	// Disable shaders
	m_pd3dDevice->SetVertexShader( NULL );
	m_pd3dDevice->SetPixelShader ( NULL );
#elif defined(OGL2)
	// Disable shaders
	glUseProgramObjectARB( 0 );
#endif

	return 0;
} // CIntManager::RenderDebug

////////////////////////////////////////////////////////////////////////////////
/// \brief Access the off-screen render targets to retrieve the computed 
///        attributes of the region of interest.
///
/// \param pAttList An array of floats, one for each attribute value in the 
///        region of interest. For example, a 5x5 region of interest with only
///        a 3D position attribute (3 values for each attribute) demands an
///        array of 5*5*3 floats.
///
/// \return 0 on success.
///         -1 if the D3D device is invalid (D3D only).
///         -2 if CIntManager::m_bIsOK is false.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::Decode( float *pAttList[ATTTYPE_SIZE] )
{
	if( !m_bIsOK ) return -2;

	// Calculates ROI's width and height
	int iROIWidth  = m_iRight  - m_iLeft + 1;
	int iROIHeight = m_iBottom - m_iTop  + 1;

	// Counter of pAttList elements
	int iAttCt = 0;

#if defined(D3D9)
	if( m_pd3dDevice == NULL )
		return -1;

	HRESULT hrResult = S_OK;

	RECT rcRect;
	rcRect.left   = 0;
	rcRect.top    = 0;
	rcRect.right  = iROIWidth  - 1;
	rcRect.bottom = iROIHeight - 1;

	D3DLOCKED_RECT d3dLockRect;	

	int iNumCurAtts;
	ATTENCODING *pCurAtts[ATTTYPE_SIZE];

	int iAttSize[ATTTYPE_SIZE] = {0};

	if( m_bAttUserDefNI )
	{
		// Copy contents of resource on D3DPOOL_DEFAULT to 
		// D3DPOOL_SYSTEMMEM
		hrResult = m_pd3dDevice->GetRenderTargetData( m_pRTS[m_iNumRTs-1], 
													  m_pReadSurf );
		if( FAILED( hrResult ) )
			return -1;

		int iAttCt = 0;
		hrResult = m_pReadSurf->LockRect( &d3dLockRect, &rcRect, 
										  D3DLOCK_READONLY );
		if( SUCCEEDED( hrResult ) )
		{
			char *pData = (char*) d3dLockRect.pBits;

			// For each row of pixels...
			for( int iY = 0; iY < iROIHeight; iY++ )
			{
				char *pRow = pData + iY * d3dLockRect.Pitch;

				for( int iX = 0; iX < iROIWidth; iX++ )
				{
					// Get color components
					float *pPx = (float*) ( pRow+iX*4*sizeof(float) );

					pAttList[ATTTYPE_USERDEFNI][iAttCt++] = pPx[0];
					pAttList[ATTTYPE_USERDEFNI][iAttCt++] = pPx[1];
					pAttList[ATTTYPE_USERDEFNI][iAttCt++] = pPx[2];
					pAttList[ATTTYPE_USERDEFNI][iAttCt++] = pPx[3];
				}
			}

			m_pReadSurf->UnlockRect( );
		}
	}

	if( !m_bRenderIdx )
		return 0;

	// For each render target...
	int iCt = 0, iRTSum = 0;
	for( int iRP = 0; iRP < m_iNumRPs; iRP++ )
	{
		if( iRP > 0 ) iRTSum = m_pRenderPass[iRP-1].m_iRTs;

		for( int iRT = 0; iRT < m_pRenderPass[iRP].m_iRTs; iRT++ )
		{
			iNumCurAtts = 0;

			// Select the attributes with values used in this render target
			for( int iAttEnc = 0; iAttEnc < ATTTYPE_SIZE; iAttEnc++ )
			{
				if( m_pAttEncoding[iAttEnc].m_bActive &&
					m_pAttEncoding[iAttEnc].m_iRP == iRP )
				{
					for( int iV = 0; iV < m_pAttEncoding[iAttEnc].m_iNumAtts; 
						 iV++ )
					{
						if( m_pAttEncoding[iAttEnc].m_iCompRT[iV] == 
							iRT + iRTSum )
						{
							pCurAtts[iNumCurAtts++] = &m_pAttEncoding[iAttEnc];
							iV = m_pAttEncoding[iAttEnc].m_iNumAtts;
						}
					}
				}
			}

			// Copy contents of resource on D3DPOOL_DEFAULT to 
			// D3DPOOL_SYSTEMMEM
			hrResult = m_pd3dDevice->GetRenderTargetData( m_pRTS[iCt], 
														  m_pReadSurf );
			if( FAILED( hrResult ) )
				return -1;

			hrResult = m_pReadSurf->LockRect( &d3dLockRect, &rcRect, 
											  D3DLOCK_READONLY );
			if( SUCCEEDED( hrResult ) )
			{
				char *pData = (char*) d3dLockRect.pBits;

				// For each row of pixels...
				for( int iY = 0; iY < iROIHeight; iY++ )
				{
					char *pRow = pData + iY * d3dLockRect.Pitch;

					for( int iX = 0; iX < iROIWidth; iX++ )
					{
						// Get color components
						float *pPx = (float*) ( pRow+iX*4*sizeof(float) );

						int iMaxComps = 4;
						if( iRT == m_pRenderPass[iRP].m_iRTs-1 )
							iMaxComps -= m_pRenderPass[iRP].m_iComps;

						int iAttEncIni = 0;
						for( int iComp = 0; iComp < iMaxComps; )
						{
							for( int iAttEnc = iAttEncIni; 
								 iAttEnc < iNumCurAtts; iAttEnc++ )
							{
								ATTENCODING *pCurAtt = pCurAtts[iAttEnc];
								int iNumAtts = pCurAtt->m_iNumAtts;
								int iOffset = (iY*iROIWidth*iNumAtts) + 
											  (iX*iNumAtts);
								for( int iV = 0; iV < iNumAtts; iV++ )
								{
									if( pCurAtt->m_iCompRT[iV] == iRT+iRTSum &&
										pCurAtt->m_iComp[iV] == iComp )
									{
										pAttList[pCurAtt->m_eType]
											[iOffset+iV] = pPx[iComp];
										
										if( iV == iNumAtts-1 )
											iAttEncIni = iAttEnc + 1;

										goto e_ContLoop;
									}
								}
							}
e_ContLoop:
							iComp++;
						}
					}
				}
			
				m_pReadSurf->UnlockRect( );
			}

			iCt++;
		}
	}

#elif defined(OGL2)
	// For each render target...
	for( int iCt = 0; iCt < m_iNumRTs; iCt++ )
	{
		glBindTexture( GL_TEXTURE_2D, m_uiRTs[iCt] );
		glGetTexImage( GL_TEXTURE_2D, 0, GL_RGBA, GL_FLOAT, pAttList[iCt] );
	}
#endif

	return 0;
} // CIntManager::Decode

////////////////////////////////////////////////////////////////////////////////
/// \brief Create the render target textures for storing the attributes
///        processed by the architecture of interaction.
///
/// \return 0 on success.
///         -1 on error.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::CreateRTs( )
{
	if( !m_bIsOK ) return -1;

	// Calculates ROI's width and height
	int iROIWidth  = m_iRight  - m_iLeft + 1;
	int iROIHeight = m_iBottom - m_iTop  + 1;

#if defined(D3D9)
	HRESULT hrResult = S_OK;

	// For each render target...
	for( int iCt = 0; iCt < m_iNumRTs; iCt++ )
	{
		// Create render target texture
		hrResult = m_pd3dDevice->CreateTexture( iROIWidth, iROIHeight, 1,
												D3DUSAGE_RENDERTARGET,
												D3DFMT_A32B32G32R32F,
												D3DPOOL_DEFAULT,
												&m_pRTT[iCt], NULL );
		if( FAILED( hrResult ) )
			return -1;

		// Get surface from texture
		hrResult = m_pRTT[iCt]->GetSurfaceLevel( 0, &m_pRTS[iCt] );
		if( FAILED( hrResult ) )
			return -1;
	}

	// Create off-screen plain surface for reading the render target
	// to system memory
	hrResult = m_pd3dDevice->CreateOffscreenPlainSurface( 
		iROIWidth, iROIHeight, D3DFMT_A32B32G32R32F, D3DPOOL_SYSTEMMEM,
		&m_pReadSurf, NULL );
	if( FAILED( hrResult ) )
		return -1;

	// Get back buffer because we are supposed to switch to it after rendering
	// to the off-screen render targets
	hrResult = m_pd3dDevice->GetRenderTarget( 0, &m_pd3dBackBuffer );
	if( FAILED( hrResult ) )
		return -1;

#elif defined(OGL2)
	// Get maximum number of supported multiple render targets
	glGetIntegerv( GL_MAX_COLOR_ATTACHMENTS_EXT, &m_iNumMRTs );

	// Create FBO and bind it
    glGenFramebuffersEXT( 1, &m_uiFBO );
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, m_uiFBO );

	GLint vVP[4];
	glGetIntegerv( GL_VIEWPORT, vVP );
	glRenderbufferStorageEXT( GL_RENDERBUFFER_EXT,
							  GL_DEPTH_COMPONENT24,
							  vVP[2], vVP[3] );

	// Create textures
	glGenTextures( m_iNumRTs, m_uiRTs );

	// Set up textures
	for( int iCt = 0; iCt < m_iNumRTs; iCt++ )
	{
		// GL_TEXTURE_RECTANGLE_ARB
		glBindTexture  ( GL_TEXTURE_2D, m_uiRTs[iCt] ); 
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
		glTexImage2D   ( GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, 
						 iROIWidth, iROIHeight, 0, GL_RGBA, GL_FLOAT, NULL );
	}

	// Re-enable rendering to the back buffer
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
#endif

	return 0;
} // CIntManager::CreateRTs

////////////////////////////////////////////////////////////////////////////////
/// \brief Release the render target textures created with 
///        CIntManager::CreateRTs.
///
/// \return 0 on success.
///         -1 on error.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::DestroyRTs( )
{
	if( !m_bIsOK ) return -1;

#if defined(D3D9)
	// Release render target textures and surfaces
	for( int iCt = 0; iCt < m_iNumRTs; iCt++ )
	{
		SAFE_RELEASE( m_pRTS[iCt] );
		SAFE_RELEASE( m_pRTT[iCt] );
	}

	// Release the reference count to the back buffer
	SAFE_RELEASE( m_pd3dBackBuffer );

	// Release the reference count to the system memory surface used
	// for reading the render targets
	SAFE_RELEASE( m_pReadSurf );
#elif defined(OGL2)
	// Release the render target textures
	glDeleteTextures( m_iNumRTs, m_uiRTs );

	// Release the FBO
	glDeleteFramebuffersEXT( 1, &m_uiFBO );
#endif

	return 0;
} // CIntManager::DestroyRTs

////////////////////////////////////////////////////////////////////////////////
/// \brief Create the shaders.
///
/// \return 0 on success.
///         -1 on error.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::CreateShaders( )
{
#if defined(D3D9)
	HRESULT hrResult S_OK;

	// Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders 
	// with the shader debugger. Debugging vertex shaders requires either REF 
	// or software vertex processing, and debugging pixel shaders requires REF.  
	// The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug experience 
	// in the shader debugger.  It enables source level debugging, prevents 
	// instruction reordering, prevents dead code elimination, and forces the 
	// compiler to compile against the next higher available software target, 
	// which ensures that the unoptimized shaders do not exceed the shader model 
	// limitations.  Setting these flags will cause slower rendering since the 
	// shaders will be unoptimized and forced into software.  See the DirectX 
	// documentation for more information about using the shader debugger.
	DWORD dwShaderFlags = 0;
#ifdef DEBUG_VS
	dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
#endif
#ifdef DEBUG_PS
	dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
#endif

	//
	// Create TBN computation effect
	//

	// Load effect from file
	LPD3DXBUFFER pd3dBuffer = NULL;
	hrResult = D3DXCreateEffectFromFile( m_pd3dDevice, C2T( "tbn.fx" ), 
		NULL, NULL, dwShaderFlags, NULL, &m_pTBNFx, &pd3dBuffer );

	// If there are errors, print them
	if( pd3dBuffer )
		fprintf( stderr, (char *) pd3dBuffer->GetBufferPointer( ) );

	if( FAILED( hrResult ) )
		return -1;

	// Get handle to uniform parameters in the effect
	m_hTBNAdjFaceIdxC = m_pTBNFx->GetParameterByName( NULL, "g_fAdjFaceIdxC" );
	m_hTBNVtIdxC      = m_pTBNFx->GetParameterByName( NULL, "g_fVtIdxC"      );
    m_hTBNFaceIdxC    = m_pTBNFx->GetParameterByName( NULL, "g_fFaceIdxC"    );

	//
	// Create N computation effect
	//

	// Load effect from file
	hrResult = D3DXCreateEffectFromFile( m_pd3dDevice, C2T( "n.fx" ), 
		NULL, NULL, dwShaderFlags, NULL, &m_pNFx, &pd3dBuffer );

	// If there are errors, print them
	if( pd3dBuffer )
		fprintf( stderr, (char *) pd3dBuffer->GetBufferPointer( ) );

	if( FAILED( hrResult ) )
		return -1;

	// Get handle to uniform parameters in the effect
	m_hNAdjFaceIdxC = m_pNFx->GetParameterByName( NULL, "g_fAdjFaceIdxC" );
	m_hNVtIdxC      = m_pNFx->GetParameterByName( NULL, "g_fVtIdxC"      );
    m_hNFaceIdxC    = m_pNFx->GetParameterByName( NULL, "g_fFaceIdxC"    );

	//
	// Create tensor computation effect
	//

	// Load effect from file
	hrResult = D3DXCreateEffectFromFile( m_pd3dDevice, C2T( "tensor.fx" ),
		NULL, NULL, dwShaderFlags, NULL, 
		&m_pTensorFx, &pd3dBuffer );

	// If there are errors, print them
	if( pd3dBuffer )
		fprintf( stderr, (char *) pd3dBuffer->GetBufferPointer( ) );

	if( FAILED( hrResult ) )
		return -1;

	// Get handle to uniform parameters in the effect
	m_hTensorAdjVtIdxC = m_pTensorFx->GetParameterByName( 
		NULL, "g_fAdjVtIdxC" );
	m_hTensorVtIdxC    = m_pTensorFx->GetParameterByName( 
		NULL, "g_fVtIdxC"    );

#elif defined(OGL2)
	FILE *fHandle = NULL;
	GLhandleARB hSO = 0;

	// Create a temporary buffer for reading the shaders from disk
	GLcharARB *szTmp = new GLcharARB[32768];
	if( !szTmp )
		return -1;

	//
	// Create TBN (face) computation effect
	//

	// Create program object
	m_hFaceTBNPO = glCreateProgramObjectARB( );

	// Read shader from disk
	szTmp[0] = NULL;
	fHandle = fopen( "facetbn.glsl", "rt" );
	fread( szTmp, sizeof(GLcharARB), 32767, fHandle );
	fclose( fHandle );

	// Create shader objects
	hSO = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	glShaderSourceARB( hSO, 1, (const GLcharARB**) &szTmp, NULL );
	glCompileShaderARB( hSO );

	// Check compilation status
	GLint iResult = 0;
	glGetObjectParameterivARB( hSO, GL_OBJECT_COMPILE_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( hPSS0 );
		return -1;
	}

	// Attach shaders to program
	glAttachObjectARB( m_hFaceTBNPO, hSO );
	// Flag shaders for deletion when they are no longer referenced
	glDeleteObjectARB( hSO );

	glLinkProgramARB( m_hFaceTBNPO );

	// Check linkage status
	glGetObjectParameterivARB( m_hFaceTBNPO, GL_OBJECT_LINK_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( m_hVtDefPO );
		return -1;
	}

	// Use face TBN shader
	glUseProgramObjectARB( m_hFaceTBNPO );

	// Get handlers to uniform parameters
	m_iFaceTBNVtMapSize = glGetUniformLocationARB( m_hFaceTBNPO, "g_iVtMapSize" );

	// Disable program object
	glUseProgramObjectARB( 0 );

	//
	// Create TBN (vertex) computation effect
	//

	// Create program object
	m_hVertexTBNPO = glCreateProgramObjectARB( );

	// Read shader from disk
	szTmp[0] = NULL;
	fHandle = fopen( "vertextbn.glsl", "rt" );
	fread( szTmp, sizeof(GLcharARB), 32767, fHandle );
	fclose( fHandle );

	// Create shader objects
	hSO = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	glShaderSourceARB( hSO, 1, (const GLcharARB**) &szTmp, NULL );
	glCompileShaderARB( hSO );

	// Check compilation status
	iResult = 0;
	glGetObjectParameterivARB( hSO, GL_OBJECT_COMPILE_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( hPSS0 );
		return -1;
	}

	// Attach shaders to program
	glAttachObjectARB( m_hVertexTBNPO, hSO );
	// Flag shaders for deletion when they are no longer referenced
	glDeleteObjectARB( hSO );

	glLinkProgramARB( m_hVertexTBNPO );

	// Check linkage status
	glGetObjectParameterivARB( m_hVertexTBNPO, GL_OBJECT_LINK_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( m_hVtDefPO );
		return -1;
	}

	// Use vertex TBN shader
	glUseProgramObjectARB( m_hVertexTBNPO );

	// Get handlers to uniform parameters
	m_iVertexTBNAdjFaceMapSize = glGetUniformLocationARB( m_hVertexTBNPO, "g_iAdjFaceMapSize" );
	m_iVertexTBNFaceMapSize    = glGetUniformLocationARB( m_hVertexTBNPO, "g_iFaceMapSize"    );

	// Disable program object
	glUseProgramObjectARB( 0 );

	//
	// Create N (face) computation effect
	//

	// Create program object
	m_hFaceNPO = glCreateProgramObjectARB( );

	// Read shader from disk
	szTmp[0] = NULL;
	fHandle = fopen( "facen.glsl", "rt" );
	fread( szTmp, sizeof(GLcharARB), 32767, fHandle );
	fclose( fHandle );

	// Create shader objects
	hSO = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	glShaderSourceARB( hSO, 1, (const GLcharARB**) &szTmp, NULL );
	glCompileShaderARB( hSO );

	// Check compilation status
	iResult = 0;
	glGetObjectParameterivARB( hSO, GL_OBJECT_COMPILE_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( hPSS0 );
		return -1;
	}

	// Attach shaders to program
	glAttachObjectARB( m_hFaceNPO, hSO );
	// Flag shaders for deletion when they are no longer referenced
	glDeleteObjectARB( hSO );

	glLinkProgramARB( m_hFaceNPO );

	// Check linkage status
	glGetObjectParameterivARB( m_hFaceNPO, GL_OBJECT_LINK_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( m_hVtDefPO );
		return -1;
	}

	// Use N (face) shader
	glUseProgramObjectARB( m_hFaceNPO );

	// Get handlers to uniform parameters
	m_iFaceNVtMapSize = glGetUniformLocationARB( m_hFaceNPO, "g_iVtMapSize" );

	// Disable program object
	glUseProgramObjectARB( 0 );

	//
	// Create N (vertex) computation shader
	//

	// Create program object
	m_hVertexNPO = glCreateProgramObjectARB( );

	// Read shader from disk
	szTmp[0] = NULL;
	fHandle = fopen( "vertexn.glsl", "rt" );
	fread( szTmp, sizeof(GLcharARB), 32767, fHandle );
	fclose( fHandle );

	// Create shader objects
	hSO = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	glShaderSourceARB( hSO, 1, (const GLcharARB**) &szTmp, NULL );
	glCompileShaderARB( hSO );

	// Check compilation status
	iResult = 0;
	glGetObjectParameterivARB( hSO, GL_OBJECT_COMPILE_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( hPSS0 );
		return -1;
	}

	// Attach shaders to program
	glAttachObjectARB( m_hVertexNPO, hSO );
	// Flag shaders for deletion when they are no longer referenced
	glDeleteObjectARB( hSO );

	glLinkProgramARB( m_hVertexNPO );

	// Check linkage status
	glGetObjectParameterivARB( m_hVertexNPO, GL_OBJECT_LINK_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( m_hVtDefPO );
		return -1;
	}

	// Use N (vertex) shader
	glUseProgramObjectARB( m_hVertexNPO );

	// Get handlers to uniform parameters
	m_iVertexNAdjFaceMapSize = glGetUniformLocationARB( m_hVertexNPO, "g_iAdjFaceMapSize" );

	m_iVertexNFaceMapSize = glGetUniformLocationARB( m_hVertexNPO, "g_iVtMapSize"   );

	// Disable program object
	glUseProgramObjectARB( 0 );

	//
	// Create 2nd order tensor computation shader
	//

	// Create program object
	m_h2ndTensorPO = glCreateProgramObjectARB( );

	// Read shader from disk
	szTmp[0] = NULL;
	fHandle = fopen( "tensor2.glsl", "rt" );
	fread( szTmp, sizeof(GLcharARB), 32767, fHandle );
	fclose( fHandle );

	// Create shader objects
	hSO = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	glShaderSourceARB( hSO, 1, (const GLcharARB**) &szTmp, NULL );
	glCompileShaderARB( hSO );

	// Check compilation status
	iResult = 0;
	glGetObjectParameterivARB( hSO, GL_OBJECT_COMPILE_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( hPSS0 );
		return -1;
	}

	// Attach shaders to program
	glAttachObjectARB( m_h2ndTensorPO, hSO );
	// Flag shaders for deletion when they are no longer referenced
	glDeleteObjectARB( hSO );

	glLinkProgramARB( m_h2ndTensorPO );

	// Check linkage status
	glGetObjectParameterivARB( m_h2ndTensorPO, GL_OBJECT_LINK_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( m_hVtDefPO );
		return -1;
	}

	// Use 2nd order tensor shader
	glUseProgramObjectARB( m_h2ndTensorPO );

	// Get handlers to uniform parameters
	m_i2ndTensorAdjVtMapSize = glGetUniformLocationARB( m_h2ndTensorPO, "g_iAdjVtMapSize" );
	m_i2ndTensorVtMapSize    = glGetUniformLocationARB( m_h2ndTensorPO, "g_iVtMapSize"    );

	// Disable program object
	glUseProgramObjectARB( 0 );

	//
	// Create 3rd order tensor computation shader
	//

	// Create program object
	m_h3rdTensorPO = glCreateProgramObjectARB( );

	// Read shader from disk
	szTmp[0] = NULL;
	fHandle = fopen( "tensor3.glsl", "rt" );
	fread( szTmp, sizeof(GLcharARB), 32767, fHandle );
	fclose( fHandle );

	// Create shader objects
	hSO = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
	glShaderSourceARB( hSO, 1, (const GLcharARB**) &szTmp, NULL );
	glCompileShaderARB( hSO );

	// Check compilation status
	iResult = 0;
	glGetObjectParameterivARB( hSO, GL_OBJECT_COMPILE_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( hPSS0 );
		return -1;
	}

	// Attach shaders to program
	glAttachObjectARB( m_h3rdTensorPO, hSO );
	// Flag shaders for deletion when they are no longer referenced
	glDeleteObjectARB( hSO );

	glLinkProgramARB( m_h3rdTensorPO );

	// Check linkage status
	glGetObjectParameterivARB( m_h3rdTensorPO, GL_OBJECT_LINK_STATUS_ARB, &iResult );
	if( !iResult )
	{
		//		glGetInfoLogARB( m_hVtDefPO );
		return -1;
	}

	// Use 3rd order tensor shader
	glUseProgramObjectARB( m_h3rdTensorPO );

	// Get handlers to uniform parameters
	m_i3rdTensorAdjVtMapSize = glGetUniformLocationARB( m_h3rdTensorPO, "g_iAdjVtMapSize" );
	m_i3rdTensorVtMapSize    = glGetUniformLocationARB( m_h3rdTensorPO, "g_iVtMapSize"    );

	// Disable program object
	glUseProgramObjectARB( 0 );

	SAFE_DELETE( szTmp );
#endif

	return 0;
} // CIntManager::CreateShaders

////////////////////////////////////////////////////////////////////////////////
/// \brief Release the shaders created with CIntObj::CreateShaders.
///
/// \return 0 on success.
///         -1 on error.
////////////////////////////////////////////////////////////////////////////////
int CIntManager::DestroyShaders( )
{
#if defined(D3D9)
	SAFE_RELEASE( m_pTBNFx    );
	SAFE_RELEASE( m_pNFx      );
	SAFE_RELEASE( m_pTensorFx );
#elif defined(OGL2)
	glUseProgramObjectARB( 0 );
	glDeleteObjectARB( m_hFaceTBNPO   );
	glDeleteObjectARB( m_hVertexTBNPO );
	glDeleteObjectARB( m_hFaceNPO     );
	glDeleteObjectARB( m_hVertexNPO   );
	glDeleteObjectARB( m_h2ndTensorPO );
	glDeleteObjectARB( m_h3rdTensorPO );
#endif

	return 0;
} // CIntManager::DestroyShaders

#if defined(D3D9)

////////////////////////////////////////////////////////////////////////////////
/// \brief Initialize device dependent objects that are not lost on resets.
///        This is a D3D-only function.
///
/// The device has been created. Resources that are not lost on reset are 
/// created here -- resources in \c D3DPOOL_MANAGED,
/// \c D3DPOOL_SCRATCH, or \c D3DPOOL_SYSTEMMEM. Image surfaces created via
/// CreateImageSurface() are never lost and are created here.  Vertex
/// shaders and pixel shaders are also created here as they are not
/// lost on rest. Paired with CIntManager::DeleteDeviceObjects().
///
/// \param pd3dDevice Pointer to the D3D device interface.
///
/// \return \c S_OK on success.
////////////////////////////////////////////////////////////////////////////////
HRESULT CIntManager::InitDeviceObjects( LPDIRECT3DDEVICE9 pd3dDevice )
{
	if( !m_bIsOK ) return E_FAIL;

	HRESULT hrResult = S_OK;

	if( m_bInitDOCalled )
		return hrResult;

	m_pd3dDevice = pd3dDevice;

	// Get maximum number of supported multiple render targets
	D3DCAPS9 d3dCaps;
	hrResult = m_pd3dDevice->GetDeviceCaps( &d3dCaps );
	if( FAILED( hrResult ) )
		return hrResult;

	m_iNumMRTs = d3dCaps.NumSimultaneousRTs;

	// Call InitDeviceObjects for each created object
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		 stlIt != m_stlObjList.end( ); stlIt++ )
	{
		if( FAILED( hrResult = (*stlIt)->InitDeviceObjects( ) ) )
			return hrResult;
	}

	if( CreateShaders( ) )
		return E_FAIL;

	m_bInitDOCalled   = true;
	m_bDeleteDOCalled = false;

	return hrResult;
} // CIntManager::InitDeviceObjects

////////////////////////////////////////////////////////////////////////////////
/// \brief Restore device dependent objects that get lost after a reset.
///        This is a D3D-only function.
///
/// The device exists, but may have just been Reset. Resources in
/// \c D3DPOOL_DEFAULT and any other device state that persists during
/// rendering should be set here. Render states, matrices, textures,
/// etc., that don't change during rendering are set once here.
/// Paired with CIntManager::InvalidateDeviceObjects().
///
/// \return \c S_OK on success.
////////////////////////////////////////////////////////////////////////////////
HRESULT CIntManager::RestoreDeviceObjects( )
{
	if( !m_bIsOK ) return E_FAIL;

	HRESULT hrResult = S_OK;

	if( m_bRestoreDOCalled )
		return hrResult;

	// Create render targets
	if( CreateRTs( ) )
		return E_FAIL;

	// Call RestoreDeviceObjects for each created object
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		stlIt != m_stlObjList.end( ); stlIt++ )
	{
		if( FAILED( hrResult = (*stlIt)->RestoreDeviceObjects( ) ) )
			return hrResult;
	}

	if( m_pTBNFx )
		m_pTBNFx->OnResetDevice( );

	if( m_pNFx )
		m_pNFx->OnResetDevice( );

	if( m_pTensorFx )
		m_pTensorFx->OnResetDevice( );

	m_bRestoreDOCalled    = true;
	m_bInvalidateDOCalled = false;

	return hrResult;
} // CIntManager::RestoreDeviceObjects

////////////////////////////////////////////////////////////////////////////////
/// \brief Invalidate device objects. This is a D3D-only function.
///
/// Invalidate device objects initialized by 
/// CIntManager::RestoreDeviceObjects().
/// Paired with CIntManager::RestoreDeviceObjects().
///
/// \return \c S_OK on success.
////////////////////////////////////////////////////////////////////////////////
HRESULT CIntManager::InvalidateDeviceObjects( )
{
	if( !m_bIsOK ) return E_FAIL;

	HRESULT hrResult = S_OK;

	if( m_bInvalidateDOCalled )
		return hrResult;

	if( DestroyRTs( ) )
		return E_FAIL;

	// Call InvalidateDeviceObjects for each created object
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		stlIt != m_stlObjList.end( ); stlIt++ )
	{
		if( FAILED( hrResult = (*stlIt)->InvalidateDeviceObjects( ) ) )
			return hrResult;
	}

	if( m_pTBNFx )
		m_pTBNFx->OnLostDevice( );

	if( m_pNFx )
		m_pNFx->OnLostDevice( );

	if( m_pTensorFx )
		m_pTensorFx->OnLostDevice( );

	m_bInvalidateDOCalled = true;
	m_bRestoreDOCalled    = false;

	return hrResult;
} // CIntManager::InvalidateDeviceObjects

////////////////////////////////////////////////////////////////////////////////
/// \brief Delete device objects. This is a D3D-only function.
///
/// Called when the manager is exiting, or the device is being changed,
/// this function deletes any device dependent objects.
/// Paired with CIntManager::InitDeviceObjects().
///
/// \return \c S_OK on success.
////////////////////////////////////////////////////////////////////////////////
HRESULT CIntManager::DeleteDeviceObjects( )
{
	if( !m_bIsOK ) return E_FAIL;

	HRESULT hrResult = S_OK;

	if( m_bDeleteDOCalled )
		return hrResult;

	// Call DeleteDeviceObjects for each created object
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		stlIt != m_stlObjList.end( ); stlIt++ )
	{
		if( FAILED( hrResult = (*stlIt)->DeleteDeviceObjects( ) ) )
			return hrResult;
	}

	if( DestroyShaders( ) )
		return E_FAIL;

	m_bDeleteDOCalled = true;
	m_bInitDOCalled   = false;

	return hrResult;
} // CIntManager::DeleteDeviceObjects

#endif

////////////////////////////////////////////////////////////////////////////////
/// \brief Computes the projection matrix for enclosing rendering to the 
///        current region of interest.
///
/// \return None.
////////////////////////////////////////////////////////////////////////////////
void CIntManager::ComputeROIMatrix( )
{
#if defined(D3D9)
	// Get current viewport settings
	D3DVIEWPORT9 d3dVP;
	m_pd3dDevice->GetViewport( &d3dVP );

	// Get current projection matrix (hope this is not a pure device!)
	D3DXMATRIX mP;
	m_pd3dDevice->GetTransform( D3DTS_PROJECTION, &mP );

	// Create an identity matrix
	D3DXMATRIX mI;
	D3DXMatrixIdentity( &mI );

	// Convert top-left and bottom-right frustum locations from window space
	// to camera space
	D3DXVECTOR3 vCamCorner1, vCamCorner2;
	D3DXVECTOR3 vWndCorner1 = 
		D3DXVECTOR3( (float) m_iLeft,  (float) m_iTop,    0.0f );
	D3DXVECTOR3 vWndCorner2 = 
		D3DXVECTOR3( (float) m_iRight + 1, (float) m_iBottom + 1, 0.0f );
	D3DXVec3Unproject( &vCamCorner1, &vWndCorner1, &d3dVP, &mP, &mI, &mI );
	D3DXVec3Unproject( &vCamCorner2, &vWndCorner2, &d3dVP, &mP, &mI, &mI );

	// Compute the new projection matrix for rendering to the
	// region of interest
	D3DXMatrixPerspectiveOffCenterLH( &m_mROIProj, 
		                              vCamCorner1.x, // Left
		                              vCamCorner2.x, // Right
		                              vCamCorner2.y, // Bottom
		                              vCamCorner1.y, // Top
		                              m_fZNear, m_fZFar );
#elif defined(OGL2)
	// Get current viewport settings
	GLint vVP[4];
	glGetIntegerv( GL_VIEWPORT, vVP );

	// Get current projection matrix
	GLdouble mP[16];
	glGetDoublev( GL_PROJECTION_MATRIX, mP );

	// Create an identity matrix
	GLdouble mI[16] = { 0.0 };
	mI[0] = mI[5] = mI[10] = mI[15] = 1.0;

	// Convert top-left and bottom-right frustum locations from window space
	// to camera space
	GLdouble vCamCorner1[3], vCamCorner2[3];
	GLdouble vWndCorner1[3] = 
		{ (double) m_iLeft,  (double) vVP[3] - (double) m_iTop,    0.0 };
	GLdouble vWndCorner2[3] = 
		{ (double) m_iRight+1, (double) vVP[3] - (double) m_iBottom+1, 0.0 };
	gluUnProject( vWndCorner1[0], vWndCorner1[1], vWndCorner1[2], mI, mP, vVP, 
				  &vCamCorner1[0], &vCamCorner1[1], &vCamCorner1[2] );
	gluUnProject( vWndCorner2[0], vWndCorner2[1], vWndCorner2[2], mI, mP, vVP, 
				  &vCamCorner2[0], &vCamCorner2[1], &vCamCorner2[2] );

	// Compute the new projection matrix for rendering to the
	// region of interest
	glMatrixMode( GL_PROJECTION );
	glPushMatrix( );
	glLoadIdentity( );
	glFrustum( vCamCorner1[0], vCamCorner2[0], 
			   vCamCorner2[1], vCamCorner1[1], m_fZNear, m_fZFar );
	glGetFloatv( GL_PROJECTION_MATRIX, m_mROIProj );
	glPopMatrix( );
#endif
} // CIntManager::ComputeROIMatrix

////////////////////////////////////////////////////////////////////////////////
/// \brief Debug function to show the contents of the render targets.
///
/// \return None.
////////////////////////////////////////////////////////////////////////////////
void CIntManager::ShowRT( )
{
	for( stlObjListIt stlIt = m_stlObjList.begin( );
		stlIt != m_stlObjList.end( ); stlIt++ )
	{
		(*stlIt)->ShowRT( );
	}
} // CIntManager::ShowRT

////////////////////////////////////////////////////////////////////////////////
/// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
///
/// Permission is hereby granted, free of charge, to any person obtaining
/// a copy of this software and associated documentation files (the 
/// "Software"), to deal in the Software without restriction, including
/// without limitation the rights to use, copy, modify, merge, publish,
/// distribute, sublicense, and/or sell copies of the Software, and to
/// permit persons to whom the Software is furnished to do so, subject to
/// the following conditions:
///
/// The above copyright notice and this permission notice shall be
/// included in all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
/// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
/// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
/// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
/// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
/// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////
