///////////////////////////////////////////////////////////////////////////////
// $RCSfile:$ $Revision:$
// $Source:$
// $Date:$ $Author:$ $State:$
////////////////////////////////////////////////////////////////////////////////
/// \file CIntManager.h 
/// \brief Declaration of CIntManager methods and related variables.
///
/// Header file for the manager class of the architecture of interaction.
///
/// \author Harlen Costa Batagelo (mailto:harlen@dca.fee.unicamp.br)
///
/// See Copyright Notice at the end of the file.
///
/// $Id:$
////////////////////////////////////////////////////////////////////////////////

#ifndef _CINTMANAGER_H
#define _CINTMANAGER_H

// Uncomment one of the following lines in your code:
//#define D3D9 ///< Using Direct3D 9.0 with D3DX
#define OGL2 ///< Using OpenGL 2.0 (GLEW) with GLU

#if !defined(D3D9) && !defined(OGL2)
  #error Definition of used API (D3D9 or OGL2) is missing.
#endif

#if defined(D3D9) && defined(OGL2)
  #error Cannot use both D3D9 and OGL2 at the same time.
#endif

#if defined(D3D9)
  #include <d3dx9math.h>
#elif defined(OGL2)
  #include <gl\glew.h>
//  #include <glew.h>
#endif

#define MAX_RENDER_PASSES 16 ///< Maximum number of render passes allowed.

#include <list>

/// Pre-declaration of the interaction object class, used by CIntManager.
class CIntObj; 

/// Declares a bidirectional linked list of pointers to CIntObj.
typedef std::list<CIntObj*> stlObjList;

/// Declares a forward iterator of stlObjList.
typedef std::list<CIntObj*>::iterator stlObjListIt;

////////////////////////////////////////////////////////////////////////////////
/// \name Helper macros for resource releasing and cleaning.
/// \{

#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }
#define SAFE_RELEASE(p)      { if(p) { (p)->Release(); (p)=NULL; } }

/// \}
////////////////////////////////////////////////////////////////////////////////

/// Enumeration of the supported semantics of input vertex attributes 
/// of the original geometry (input parameters of the vertex shader).
typedef enum VSSEMANTIC
{
	VSSEMANTIC_NORMAL    = 1,    ///< Input param: vNormal (vec3).
	VSSEMANTIC_COLOR0    = 2,    ///< Input param: vColor0 (vec4).
	VSSEMANTIC_COLOR1    = 3,    ///< Input param: vColor1 (vec4).
	VSSEMANTIC_TEXCOORD0 = 4,    ///< Input param: vTex0   (vec4).
	VSSEMANTIC_TEXCOORD1 = 5,    ///< Input param: vTex1   (vec4).
	VSSEMANTIC_TEXCOORD2 = 6,    ///< Input param: vTex2   (vec4).
	VSSEMANTIC_TEXCOORD3 = 7,    ///< Input param: vTex3   (vec4).
	VSSEMANTIC_TEXCOORD4 = 8,    ///< Input param: vTex4   (vec4).
	VSSEMANTIC_TEXCOORD5 = 9,    ///< Input param: vTex5   (vec4).
	VSSEMANTIC_TEXCOORD6 = 10,   ///< Input param: vTex6   (vec4).
	VSSEMANTIC_TEXCOORD7 = 11,   ///< Input param: vTex7   (vec4).
	VSSEMANTIC_MAXVALUE  = 12,   ///< Not used.
	VSSEMANTIC_END       = 0xFF,
} VSSEMANTIC, *LPVSSEMANTIC;

/// Enumeration of the supported semantic bindings between semantics
/// of input vertex attributes and output vertex attributes.
typedef enum INTSEMANTIC
{
	INTSEMANTIC_TEXCOORD  = 0,  ///< Texcoords used for TBN frame            (vec2).
	INTSEMANTIC_VERTEXID  = 1,  ///< Vertex ID                               (vec1).
	INTSEMANTIC_USERDEFI  = 2,  ///< User-def value for indexed geometry     (vec4).
	INTSEMANTIC_USERDEFNI = 3,  ///< User-def value for non-indexed geometry (vec4).
	INTSEMANTIC_MAXVALUE  = 4,  ///< Not used.
	INTSEMANTIC_END       = 0xFF,
} INTSEMANTIC, *LPINTSEMANTIC;

/// Structure used to define the binding between vertex attribute
/// semantics of the original geometry and vertex attribute semantics
/// of the architecture of interaction, i.e., how the architecture of
/// interaction interprets the input vertex attributes.
typedef struct SEMANTICBINDING
{
	/// Semantic of an input attribute of the original geometry.
	VSSEMANTIC  iVSSemantic;  
	/// Semantic of an input attribute of the pixel deformation shader.
	INTSEMANTIC iIntSemantic; 
} SEMANTICBINDING, *LPSEMANTICBINDING;

/// Defines the end of an array of SEMANTICBINDINGs
#define SEMANTIC_END { VSSEMANTIC_END, INTSEMANTIC_END }

typedef enum ATTTYPE
{
	ATTTYPE_DEPTH     = 0, ///< Depth in NDC.
	ATTTYPE_TEXCOORD  = 1, ///< Texture coordinates (UV).
	ATTTYPE_TBN       = 2, ///< TBN frame.
	ATTTYPE_USERDEFI  = 3, ///< User-defined value for indexed geometry.
	ATTTYPE_USERDEFNI = 4, ///< User-defined value for non-indexed geometry.
	ATTTYPE_NORMAL    = 5, ///< Normal vector.
	ATTTYPE_CURV      = 6, ///< Principle curvatures and directions.
	ATTTYPE_TENSOR2   = 7, ///< 2nd fundamental form (e, f, g).
	ATTTYPE_TENSOR3   = 8, ///< Tensor of curvature derivative (a, b, c, d).
	ATTTYPE_SIZE      = 9, ///< Dummy value that indicates the enum size.
	ATTTYPE_END       = 0xFF,
} ATTTYPE, *LPATTTYPE;

/// Structure used to define the attributes to be obtained from the rendered
/// geometry, in a per-pixel basis.
typedef struct ATTRIBUTE
{
	/// Per-pixel attribute to be obtained from the rendered geometry.
	ATTTYPE iAttribute;
} ATTRIBUTE, *LPATTRIBUTE;

////////////////////////////////////////////////////////////////////////////////
/// \brief Manager class of the architecture of interaction.
///
/// Provides the generic functionality needed for setting up the objects under
/// interaction and for processing the desired attributes.
////////////////////////////////////////////////////////////////////////////////
class CIntManager
{
	/// Grant private member access to CIntObj.
	friend class CIntObj;

	////////////////////////////////////////////////////////////////////////////
	/// \brief Properties of a render pass used for rendering the geometry 
	/// attributes.
	////////////////////////////////////////////////////////////////////////////
	typedef struct RENDERPASS
	{
		int m_iRP;  ///< Number of this render pass (starting from 0).
		int m_iRTs; ///< Number of render targets used in this render pass.

		/// Color components not used in the last render target of 
		/// this render pass.
		int m_iComps; 
	} RENDERPASS, *LPRENDERPASS;

	/// Array of data structures with the properties of each render pass.
	RENDERPASS m_pRenderPass[MAX_RENDER_PASSES];

	////////////////////////////////////////////////////////////////////////////
	/// \brief Data structure that contains information about the mapping 
	/// between the values of an attribute (in a maximum of 3 values per 
	/// attribute) and the color components of the output render targets.
	////////////////////////////////////////////////////////////////////////////
	typedef struct ATTENCODING
	{
		/// Whether this attribute should be calculated.
		bool m_bActive;
		/// Type of attribute (see ATTTYPE).
		int m_eType;
		/// Render pass (starting from 0).
		int m_iRP;      
		/// Which color component stores each value of an attribute. 
		/// This is a value between 0 and 3 (0123=RGBA).
		int m_iComp[16];   
		/// Which render target stores each value of an attribute. 
		/// This is a value between 0 and m_iNumRTs - 1.
		int m_iCompRT[16];
		/// Number of values stored by this attribute.
		int m_iNumAtts;
	} ATTENCODING, *LPATTENCODING;

	/// Array of data structures with the mapping between the attributes'
	/// values and the color components of the render targets.
	ATTENCODING m_pAttEncoding[ATTTYPE_SIZE];

	/// Whether the architecture of interaction is working properly.
	/// If this is false, an error has occurred and the class methods will
	/// not work. Use CIntManager::IsOK( ) to check this variable.
	bool m_bIsOK;

	/// Array of semantic bindings used by this object.
	LPSEMANTICBINDING m_pBindings;

	/// Pointer to the string containing the definition of the VSDEFORM
	/// data structure.
	char *m_szVSDeform;

	/// Pointer to the string containing the definition of the PSDEFORM
	/// data structure.
	char *m_szPSDeform;

#if defined(OGL2)
	/// Pointer to the string containing the statements for defining the 
	/// input of the VSDEFORM structure in the GLSL vertex shader.
	char *m_szInputVSDeform;
	/// Pointer to the string containing the statements for defining the 
	/// input of the PSDEFORM structure in the GLSL fragment shader.
	char *m_szInputPSDeform;
	/// Pointer to the string containing the statements for returning the
	/// contents of PSDEFORM in a GLSL vertex shader.
	char *m_szReturnPSDeform;
#endif

	/// List of objects under interaction.
	stlObjList m_stlObjList;

	////////////////////////////////////////////////////////////////////////////
	/// \name Settings of the region of interest (ROI)
	/// \{

	int m_iLeft;   ///< X-coordinate of the upper-left corner of the ROI.
	int m_iTop;    ///< Y-coordinate of the upper-left corner of the ROI.
	int m_iRight;  ///< X-coordinate of the lower-right corner of the ROI.
	int m_iBottom; ///< Y-coordinate of the lower-right corner of the ROI. 

	/// \}
	////////////////////////////////////////////////////////////////////////////

	/// Number of multiple render targets supported.
	int m_iNumMRTs;

	/// Number of allocated render targets (this depends on the number of
	/// attributes wanted).
	int m_iNumRTs;

	/// Number of render passes needed for rendering the attributes.
	int m_iNumRPs;

	float m_fZNear; ///< Z-value of the near clipping plane.
	float m_fZFar;  ///< Z-value of the far clipping plane.

public:

	////////////////////////////////////////////////////////////////////////////
	/// \name Attribute flags.
	/// \{

	/// Whether any of the attributes of indexed geometry is being used.
	bool m_bRenderIdx;

	bool m_bAttDepth;     ///< Compute depth in NDC.
	bool m_bAttTexCoord;  ///< Compute texture coordinates.
	bool m_bAttTBN;       ///< Compute the TBN frame.
	bool m_bAttUserDefI;  ///< Compute user-def value for indexed geometry.
	bool m_bAttUserDefNI; ///< Compute user-def value for non-indexed geometry.
	bool m_bAttNormal;    ///< Compute the normal vector.
	bool m_bAttCurv;      ///< Compute curvature.
	bool m_bAttTensor2;   ///< Compute the 2nd fundamental form.
	bool m_bAttTensor3;   ///< Compute the 3rd fundamental form.

	/// \}
	////////////////////////////////////////////////////////////////////////////

	/// Maximum number of cache textures used for holding the results of the
	/// vertex-level computations on the deformed geometry.
	int m_iMaxVtCacheTex;

#if defined(D3D9)
	LPDIRECT3DDEVICE9 m_pd3dDevice; ///< Pointer to the D3D device interface.

	/// Whether CIntManager::InitDeviceObjects was called after the 
	/// last call to CIntManager::DeleteDeviceObjects.
	bool m_bInitDOCalled;

	/// Whether CIntManager::RestoreDeviceObjects was called after the 
	/// last call to CIntManager::InvalidateDeviceObjects.
	bool m_bRestoreDOCalled;

	/// Whether CIntManager::InvalidateDeviceObjects was called after the 
	/// last call to CIntManager::RestoreDeviceObjects.
	bool m_bInvalidateDOCalled;

	/// Whether CIntManager::DeleteDeviceObjects was called after the 
	/// last call to CIntManager::InitDeviceObjects.
	bool m_bDeleteDOCalled;

	/// Render target texture for rendering the geometry attributes.
	LPDIRECT3DTEXTURE9 m_pRTT[16];

	/// Surface from the render target (D3DPOOL_DEFAULT).
	LPDIRECT3DSURFACE9 m_pRTS[16];

	/// System-memory surface used to hold a copy of the current 
	/// render target being read.
	LPDIRECT3DSURFACE9 m_pReadSurf;

	/// A copy of the back buffer, used for restoring the rendering
	/// state after rendering to the off-screen surfaces.
	LPDIRECT3DSURFACE9 m_pd3dBackBuffer;

	/// Projection matrix used when rendering to the ROI (OpenGL).
	D3DXMATRIX m_mROIProj;

	/// D3DX effect interface to TBN computation shader.
	LPD3DXEFFECT m_pTBNFx;
	/// Handle to 1D-2D constant of face adjacency texture.
	D3DXHANDLE m_hTBNAdjFaceIdxC;
	/// Handle to 1D-2D constant of vertex attributes texture.
	D3DXHANDLE m_hTBNVtIdxC;
	/// Handle to 1D-2D constant of per-face vertex indices texture.
    D3DXHANDLE m_hTBNFaceIdxC;

	/// D3DX effect interface to normal computation shader.
	LPD3DXEFFECT m_pNFx;
	/// Handle to 1D-2D constant of face adjacency texture.
	D3DXHANDLE m_hNAdjFaceIdxC;
	/// Handle to 1D-2D constant of vertex attributes texture.
	D3DXHANDLE m_hNVtIdxC;
	/// Handle to 1D-2D constant of per-face vertex indices texture.
    D3DXHANDLE m_hNFaceIdxC;

	/// D3DX effect interface to tensor computation shader.
	LPD3DXEFFECT m_pTensorFx;
	/// Handle to 1D-2D constant of vertex adjacency texture.
	D3DXHANDLE m_hTensorAdjVtIdxC;
	/// Handle to 1D-2D constant of vertex attributes texture.
	D3DXHANDLE m_hTensorVtIdxC;

#elif defined(OGL2)
	/// ID of the frame buffer object.
	GLuint m_uiFBO;

	/// Symbolic constants specifying the buffers into which fragment 
	/// values will be written.
 	GLenum m_eRTs[16];

	/// IDs of the render target textures.
	GLuint m_uiRTs[16];

	/// Projection matrix used when rendering to the ROI (D3D).
	GLfloat m_mROIProj[16];

	/// Handle of the program object of the TBN computation shader.
	/// (per-face)
	GLhandleARB m_hFaceTBNPO;
	/// Handle to size of the vertex attributes texture.
	GLint m_iFaceTBNVtMapSize;

	/// Handle of the program object of the TBN computation shader.
	/// (per-vertex)
	GLhandleARB m_hVertexTBNPO;
	/// Handle to the size of the face adjacency texture.
	GLint m_iVertexTBNAdjFaceMapSize;
	/// Handle to the size of the per-face vertex indices texture.
	GLint m_iVertexTBNFaceMapSize;

	/// Handle of the program object of the normal computation shader 
	/// (per-face).
	GLhandleARB m_hFaceNPO;
	/// Handle to the size of the vertex attributes texture.
	GLint m_iFaceNVtMapSize;

	/// Handle of the program object of the normal computation shader 
	/// (per-vertex).
	GLhandleARB m_hVertexNPO;
	/// Handle to the size of the face adjacency texture.
	GLint m_iVertexNAdjFaceMapSize;
	/// Handle to the size of the per-face vertex indices texture.
	GLint m_iVertexNFaceMapSize;

	/// Handle of the program object of the 2nd order tensor computation shader
	GLhandleARB m_h2ndTensorPO;
	/// Handle to 1D-2D constant of vertex adjacency texture.
	GLint m_i2ndTensorAdjVtMapSize;
	/// Handle to 1D-2D constant of vertex attributes texture.
	GLint m_i2ndTensorVtMapSize;

	/// Handle of the program object of the 3rd order tensor computation shader
	GLhandleARB m_h3rdTensorPO;
	/// Handle to 1D-2D constant of vertex adjacency texture.
	GLint m_i3rdTensorAdjVtMapSize;
	/// Handle to 1D-2D constant of vertex attributes texture.
	GLint m_i3rdTensorVtMapSize;
#endif

	void ComputeROIMatrix( );
	int  CreateRTs       ( );
	int  DestroyRTs      ( );
	int  CreateShaders   ( );
	int  DestroyShaders  ( );

public:

	/// \brief Check if there were any errors during the execution of the
	///        architecture of interaction.
	/// \return \c true on success. \c false on error.
	inline bool IsOK( ) { return m_bIsOK; }

	////////////////////////////////////////////////////////////////////////////
	/// \name Functions related to the region of interest.
	/// \{

	int  SetROI  ( int  iLeft, int  iTop, int  iRight, int  iBottom );
	void GetROI  ( int &iLeft, int &iTop, int &iRight, int &iBottom );

	void SetZNear( float fZNear );
	void SetZFar ( float fZFar  );

	/// \brief Retrieve the current z-value of the near clipping plane.
	/// \return Current z-value of the near clipping plane.
	float GetZNear( ) { return m_fZNear; };

	/// \brief Retrieve the current z-value of the far clipping plane.
	/// \return Current z-value of the far clipping plane.
	float GetZFar ( ) { return m_fZFar; };

	/// \}
	////////////////////////////////////////////////////////////////////////////

	////////////////////////////////////////////////////////////////////////////
	/// \name Functions related to the management of objects under interaction.
	/// \{

	int  CreateObject ( CIntObj **pObj, 
						const char *szAdjWriteFile = NULL, 
						const char *szAdjReadFile  = NULL );
	int  DestroyObject( CIntObj *pObj );
	int  GetNumObjects( );
	void GetObjects   ( CIntObj **pObjList );

	/// \}
	////////////////////////////////////////////////////////////////////////////

	int Render( );
	int RenderDebug( bool bOvlWire, float fAspect, float fZNear, float fZFar,
#if defined(D3D9)
					 LPDIRECT3DTEXTURE9 pd3dActualTex,
					 LPDIRECT3DTEXTURE9 pd3dReliefTex,
					 LPDIRECT3DTEXTURE9 pd3dBumpTex,
#elif defined(OGL2)
					 GLuint uiActualTex, GLuint uiReliefTex, GLuint uidBumpTex,
#endif
					 bool bUseReliefMap, bool bUseBumpMap, int iShowAtt );


	int Decode( float *pAttList[ATTTYPE_SIZE] );

	int SetAttributes( LPATTRIBUTE pAttributes );
	int BindSemantics( LPSEMANTICBINDING pBindings );

	// Debug only
	void ShowRT( );

#if defined(D3D9)
	HRESULT InitDeviceObjects      ( LPDIRECT3DDEVICE9 pd3dDevice );
	HRESULT RestoreDeviceObjects   ( );
	HRESULT InvalidateDeviceObjects( );
	HRESULT DeleteDeviceObjects    ( );
#endif

	 CIntManager( );
	~CIntManager( );
};

#include "CIntObj.h"

#endif

////////////////////////////////////////////////////////////////////////////////
/// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
///
/// Permission is hereby granted, free of charge, to any person obtaining
/// a copy of this software and associated documentation files (the 
/// "Software"), to deal in the Software without restriction, including
/// without limitation the rights to use, copy, modify, merge, publish,
/// distribute, sublicense, and/or sell copies of the Software, and to
/// permit persons to whom the Software is furnished to do so, subject to
/// the following conditions:
///
/// The above copyright notice and this permission notice shall be
/// included in all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
/// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
/// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
/// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
/// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
/// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////