////////////////////////////////////////////////////////////////////////////////
// $RCSfile:$ $Revision:$
// $Source:$
// $Date:$ $Author:$ $State:$
////////////////////////////////////////////////////////////////////////////////
/// \file CIntObj.h 
/// \brief Declaration of CIntObj methods and related variables.
///
/// This class encapsulates data of an object (set of primitives) under
/// interaction. Instances of this class are managed by CIntManager.
///
/// \author Harlen Costa Batagelo (mailto:harlen@dca.fee.unicamp.br)
///
/// See Copyright Notice at the end of the file.
///
/// $Id:$
////////////////////////////////////////////////////////////////////////////////

#ifndef _CINTOBJ_H
#define _CINTOBJ_H

#include "CIntManager.h"

#if defined(D3D9)
/// Vertex format with transformed coordinates and a set of
/// texture coordinates for quad rendering.
#define FVF_TLVERTEX ( D3DFVF_XYZRHW | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0) )

/// Definition of a transformed vertex for quad rendering.
struct TLVERTEX
{
	D3DXVECTOR4 p; ///< Transformed position.
    D3DXVECTOR2 t; ///< 2D set of texture coordinates.
};
#endif

////////////////////////////////////////////////////////////////////////////////
/// \brief Object class of the architecture of interaction.
///
/// Encapsulates data of an object (set of primitives) under interaction and
/// managed by CIntManager.
////////////////////////////////////////////////////////////////////////////////
class CIntObj
{
	/// Grant private member access to CIntManager.
	friend class CIntManager;

	/// Pointer to the manager class.
	CIntManager *m_pManager;

	/// Whether this object is rendered in CIntObj::Render.
	/// Default is true.
	bool m_bVisible;

	////////////////////////////////////////////////////////////////////////////
	/// \name Variables related to deformation functions.
	/// \{

	/// Whether vertex-level deformation before estimation of diff. geom. 
	/// properties is used. Default is false.
	bool m_bUsePreDefVS;

	/// Whether vertex-level deformation after estimation of diff. geom. 
	/// properties is used. Default is false.
	bool m_bUsePostDefVS;

	/// When CIntObj::m_bUseVtDef is true, this defines whether the 
	/// vertex-level deformation function changes the value of the texture
	/// coordinates. By default this is false, meaning that only the position 
	/// attribute is deformed.
	bool m_bDefTex;

	/// Pointer to the string containing the vertex-level deformation shader 
	/// used before the estimation of diff. geom. properties.
	char *m_szPreDefVS;

	/// Pointer to the string containing the pixel-level deformation shader.
	char *m_szDefPS;

	/// Pointer to the string containing the vertex-level deformation shader 
	/// used after the estimation of diff. geom. properties.
	char *m_szPostDefVS;

	/// Whether fragment-level deformation is used. Default is false.
	bool m_bUsePxDef;

	/// Number of vertices in CIntObj::m_pVtPos and CIntObj::m_pVtTexCoord.
	int m_iNumVt;

	/// Pointer to an array of 3D coordinates containing the position
	/// of each vertex of the original geometry, in object space. The array 
	/// should be packed like [x1, y1, z1, x2, y2, z2, ..., xn, yn, zn]
	float *m_pVtPos;

	/// Pointer to an array of 2D texture coordinates of each vertex of the
	/// original geometry. The array should be packed like [u1, v1, u2, v2, 
	/// ..., un, vn]
	float *m_pVtTexCoord;

	/// Pointer to an array of 3D coordinates containing the normal vector
	/// of each vertex of the original geometry, in object space. The array 
	/// should be packed like [x1, y1, z1, x2, y2, z2, ..., xn, yn, zn]
	float *m_pVtNormal;

	/// Size of each dimension of the vertex attribute texture (texture
	/// of vertex position and texture coordinates).
	int m_iVtAttTexSize;

	/// Number of faces in CIntObj::m_pFaceIdx.
	int m_iNumFace;

	/// Pointer to an array of triangle face indices. Each set of 3 elements in
	/// the array contain the indices to a triangle's vertices.
	int *m_pFaceIdx;

	/// Size of each dimension of the face indices texture (texture
	/// of vertex indices of faces).
	int m_iFaceIdxTexSize;

	/// Size of the 'adjacent faces' texture.
	int m_iAdjFaceTexSize;

	/// Size of the 'adjacent vertex' texture.
	int m_iAdjVtTexSize;

	/// \}
	////////////////////////////////////////////////////////////////////////////

	/// Pointer to the user-defined data used by the callback functions.
	void *m_pUserData;

	/// Functor to callback rendering for updating vertex attributes.
	void (*m_fUpdate)( CIntObj*, void *pUserData ); 

	/// Functor to callback rendering of indexed geometry.
	void (*m_fRenderIdx)( CIntObj*, int, void *pUserData ); 

	/// Functor to callback rendering of indexed geometry.
	void (*m_fRenderNonIdx)( CIntObj*, void *pUserData ); 

	/// Name of the file where the adjacency data of the geometry
	/// will be written to. If this is NULL, the adjacency data will not
	/// be written.
	char *m_szAdjWriteFile; 

	/// Name of the file from where the adjacency data of the geometry
	/// will be read. If this is NULL, the adjacency data will not be read.
	char *m_szAdjReadFile;

#if defined(D3D9)
	/// Local transformation matrix.
	D3DXMATRIX m_mLocal;
	/// Vertices for the quad rendering on vertex attributes.
	TLVERTEX m_pScreenQuadVt[4];
	/// Vertices for the quad rendering on face attributes.
	TLVERTEX m_pScreenQuadFace[4];

	LPDIRECT3DDEVICE9  m_pd3dDevice;        ///< D3D device interface.
	LPDIRECT3DTEXTURE9 m_pd3dVtPosTex;      ///< Vertex position texture.
	LPDIRECT3DTEXTURE9 m_pd3dVtTexCoordTex; ///< Texture coordinates texture.
	LPDIRECT3DTEXTURE9 m_pd3dVtNormalTex;   ///< Vertex normal texture.
	LPDIRECT3DTEXTURE9 m_pd3dVtTanHTex;     ///< Vertex tangent and handedness.
	LPDIRECT3DTEXTURE9 m_pd3dVtTensor2Tex;  ///< 2nd tensor texture.
	LPDIRECT3DTEXTURE9 m_pd3dVtTensor3Tex;  ///< 3rd tensor texture.
	LPDIRECT3DTEXTURE9 m_pd3dVtP1K1Tex;     ///< #1 principal direction and k1.
	LPDIRECT3DTEXTURE9 m_pd3dVtP2K2Tex;     ///< #2 principal direction and k2.

	/// Array of cache textures of vertex-level computations. The size of
	/// the array is CIntManager::m_iMaxVtCacheTex.
	/// The type of attributes written to each texture will depend on the
	/// number of interaction attributes selected by the user.
	LPDIRECT3DTEXTURE9 m_pd3dVtCacheTex[4];

	LPDIRECT3DTEXTURE9 m_pd3dFaceIdxTex;    ///< Face index texture.
	LPDIRECT3DTEXTURE9 m_pd3dAdjFaceTex;    ///< 'Adjacent faces' texture.
	LPDIRECT3DTEXTURE9 m_pd3dAdjVtTex;      ///< 'Adjacent vertices' texture.
	LPDIRECT3DTEXTURE9 m_pd3dAdjAddTex;     ///< 'Adjacency addresses' texture.

	/// D3DX effect interface to the vertex-level deformation effect.
	LPD3DXEFFECT m_pVtDefFx;
	/// Handle to projection matrix in the vertex-level deformation effect.
	D3DXHANDLE m_hVtDefFxProj;
	/// Handle to variable g_iVtMapSize of the vertex-level deformation effect.
	D3DXHANDLE m_hVtDefFxVtMapSz;

	/// D3DX effect interface for rendering the geometry for visualization.
	LPD3DXEFFECT m_pRenderFx;
	/// Handle to the world matrix.
	D3DXHANDLE m_hRenderFxW;
	/// Handle to the view matrix.
	D3DXHANDLE m_hRenderFxV;
	/// Handle to the combined world*view*proj matrix of the render shader.
	D3DXHANDLE m_hRenderFxWVP;
	/// Handle to the inverse transpose of the world matrix of the 
	/// render shader.
	D3DXHANDLE m_hRenderFxITW;
	/// Handle to the variable g_fVtIdxC of the render shader.
	D3DXHANDLE m_hRenderFxIdxC;

	/// D3DX effect interface to the encoder effect using indexed geometry.
	LPD3DXEFFECT m_pEncodeIFx[MAX_RENDER_PASSES];
	/// Handle to the variable g_fVtIdxC of the encoder fx using indexed 
	/// geometry.
	D3DXHANDLE m_hEncodeIFxIdxC[MAX_RENDER_PASSES];
	/// Handle to the combined world*view*proj matrix.
	D3DXHANDLE m_hEncodeIFxWVP[MAX_RENDER_PASSES];
	/// Handle to the inverse transpose of world matrix.
	D3DXHANDLE m_hEncodeIFxITW[MAX_RENDER_PASSES];

	/// D3DX effect interface to the encoder effect using non-indexed geometry.
	LPD3DXEFFECT m_pEncodeNIFx;
	/// Handle to the variable g_fVtIdxC of the encoder fx using non-idx 
	/// geometry.
	D3DXHANDLE m_hEncodeNIFxIdxC;
	/// Handle to the combined world*view*proj matrix using non-idx geometry.
	D3DXHANDLE m_hEncodeNIFxWVP;
	/// Handle to the inverse transpose of the world matrix using 
	/// non-idx geometry.
	D3DXHANDLE m_hEncodeNIFxITW;

	/// Whether CIntManager::InitDeviceObjects was called after the 
	/// last call to CIntManager::DeleteDeviceObjects.
	bool m_bInitDOCalled;

	/// Whether CIntManager::RestoreDeviceObjects was called after the 
	/// last call to CIntManager::InvalidateDeviceObjects.
	bool m_bRestoreDOCalled;

	/// Whether CIntManager::InvalidateDeviceObjects was called after the 
	/// last call to CIntManager::RestoreDeviceObjects.
	bool m_bInvalidateDOCalled;

	/// Whether CIntManager::DeleteDeviceObjects was called after the 
	/// last call to CIntManager::InitDeviceObjects.
	bool m_bDeleteDOCalled;

	HRESULT InitDeviceObjects      ( );
	HRESULT RestoreDeviceObjects   ( );
	HRESULT InvalidateDeviceObjects( );
	HRESULT DeleteDeviceObjects    ( );
#elif defined(OGL2)
	/// Local transformation matrix.
	GLfloat m_mLocal[16];

	/// ID of the FBO used to render to vertex attribute textures.
	GLuint m_uiVtTexFBO;

	/// Array of cache textures of vertex-level computations. The size of
	/// the array is CIntManager::m_iMaxVtCacheTex.
	/// The type of attributes written to each texture will depend on the
	/// number of interaction attributes selected by the user.
	GLuint m_uiVtCacheTex[4];

	/// ID of the render target texture with deformed vertex positions.
	GLuint m_uiVtPosTexRT;

	/// ID of the render target texture with deformed vertex tex coords.
	GLuint m_uiVtTexCoordTexRT;

	/// ID of the render target texture with vertex normals.
	GLuint m_uiVtNormalTexRT;

	/// ID of the render target texture with vertex tangent and handedness.
	GLuint m_uiVtTanHTexRT;

	/// ID of the render target texture with 2nd fundamental form (per-vertex).
	GLuint m_uiVtTensor2TexRT;

	/// ID of the render target texture with tensor of curvature derivative form (per-vertex).
	GLuint m_uiVtTensor3TexRT;

	/// ID of the render target texture with p1k1.
	GLuint m_uiVtP1K1TexRT;

	/// ID of the render target texture with p2k2.
	GLuint m_uiVtP2K2TexRT;

	/// ID of the texture with face index data.
	GLuint m_uiFaceIdxTex;

	/// ID of the 'adjacent faces' texture.
	GLuint m_uiAdjFaceTex;

	/// ID of the 'adjacent vertices' texture.
	GLuint m_uiAdjVtTex;

	/// ID of the 'adjacency addresses' texture.
	GLuint m_uiAdjAddTex;

	/// Handle of the program object of the vertex-level deformation shader.
	GLhandleARB m_hVtDefPO;
	/// Handle to variable g_iVtMapSize of the vertex-level deformation effect.
	GLint m_iVtDefPOVtMapSize;

	/// Handle of the program object for rendering the geometry for visualization.
	GLhandleARB m_hRenderPO;
	/// Handle to the variable g_iVtMapSize of the render shader.
	GLint m_iRenderPOVtMapSize;

	/// Handle of the program object for encoding indexed geometry.
	GLhandleARB m_hEncodeIPO[MAX_RENDER_PASSES];
	/// Handle to variable g_iVtMapSize of the shader handled by m_hEncodeIPO.
	GLint m_iEncodeIPOVtMapSize[MAX_RENDER_PASSES];

	/// Handle of the program object for encoding non-indexed geometry.
	GLhandleARB m_hEncodeNIPO;
	/// Handle to variable g_iVtMapSize of the shader handled by m_hEncodeNIPO.
	GLint m_iEncodeNIPOVtMapSize;
#endif

	int Update        ( );
	int Render        ( );
	int RenderDebug   ( bool bOvlWire, float fAspect, 
						float fZNear, float fZFar, 
#if defined(D3D9)
					 LPDIRECT3DTEXTURE9 pd3dActualTex,
					 LPDIRECT3DTEXTURE9 pd3dReliefTex,
					 LPDIRECT3DTEXTURE9 pd3dBumpTex,
#elif defined(OGL2)
					 GLuint uiActualTex, GLuint uiReliefTex, GLuint uiBumpTex,
#endif
						bool bUseReliefMap, bool bUseBumpMap, int iShowAtt );
	int CreateRTs     ( );
	int DestroyRTs    ( );
	int CreateNRTs    ( );
	int DestroyNRTs   ( );
	int CreateShaders ( );
	int DestroyShaders( );

	// Debug only
	void ShowRT( );

#if defined(D3D9)
	bool CreatePSEncoder( int iCurRP, int iCurRT, int iCurComp, 
					      char **szShader, char *szReplace );
#elif defined(OGL2)
	bool CreatePSEncoder( int iCurRP, int iCurRT, int iCurComp, 
						  char **szVSShader, char **szPSShader, 
						  char *szReplace );
#endif


public:

#if defined(D3D9)
	void SetLocalTransform ( D3DXMATRIX mLocal );

	/// \brief Returns the interface to the effect used for deforming the
	/// geometry in the vertex level.
	const LPD3DXEFFECT GetVtDefShader( )
	{ return (const LPD3DXEFFECT) m_pVtDefFx; };

	/// \brief Returns the interface to the effect used for encoding the
	/// indexed geometry after per-vertex deformation
	/// \param iRP Render pass.
	const LPD3DXEFFECT GetEncodeIShader( int iRP )
	{ return (const LPD3DXEFFECT) m_pEncodeIFx[iRP]; }

	/// \brief Returns the interface to the effect used for encoding the
	/// non-indexed geometry after per-vertex deformation.
	const LPD3DXEFFECT GetEncodeNIShader( )
	{ return (const LPD3DXEFFECT) m_pEncodeNIFx; }

	/// \brief Returns the interface to the effect used for the 
	/// debug rendering.
	const LPD3DXEFFECT GetDebugRenderShader( )
	{ return (const LPD3DXEFFECT) m_pRenderFx; }

	/// \brief Returns the interface to the vertex position texture.
	const LPDIRECT3DTEXTURE9 GetVtPosTex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtPosTex; };
	/// \brief Returns the interface to the texture coordinates texture.
	const LPDIRECT3DTEXTURE9 GetVtTexCoordTex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtTexCoordTex; };
	/// \brief Returns the interface to the vertex normal texture.
	const LPDIRECT3DTEXTURE9 GetVtNormalTex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtNormalTex; };
	/// \brief Returns the interface to the vertex tangent and handedness.
	const LPDIRECT3DTEXTURE9 GetVtTanHTex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtTanHTex; };
	/// \brief Returns the interface to the 2nd tensor texture.
	const LPDIRECT3DTEXTURE9 GetVtTensor2Tex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtTensor2Tex; };
	/// \brief Returns the interface to the 3rd tensor texture.
	const LPDIRECT3DTEXTURE9 GetVtTensor3Tex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtTensor3Tex; };
	/// \brief Returns the interface to the #1 principal direction and k1.
	const LPDIRECT3DTEXTURE9 GetVtP1K1Tex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtP1K1Tex; };
	/// \brief Returns the interface to the#2 principal direction and k2.
	const LPDIRECT3DTEXTURE9 GetVtP2K2Tex( )
	{ return (const LPDIRECT3DTEXTURE9) m_pd3dVtP2K2Tex; };
	/// \brief Returns the current local transformation matrix
	const D3DXMATRIX GetLocalTransform( ) { return m_mLocal; };
#elif defined(OGL2)
	void SetLocalTransform( GLfloat mLocal[16] );

	/// \brief Returns the handle to the program object used for deforming the
	/// geometry in the vertex level.
	const GLhandleARB GetVtDefShader( ) 
	{ return (const GLhandleARB) m_hVtDefPO; };

	/// \brief Returns the handle to the program object used for encoding the
	/// indexed geometry after per-vertex deformation
	/// \param iRP Render pass.
	const GLhandleARB GetEncodeIShader( int iRP )
	{ return (const GLhandleARB) m_hEncodeIPO[iRP]; }

	/// \brief Returns the handle to the program object used for encoding the
	/// non-indexed geometry after per-vertex deformation.
	const GLhandleARB GetEncodeNIShader( )
	{ return (const GLhandleARB) m_hEncodeNIPO; }

	/// \brief Returns the handle to the program object used for the 
	/// debug rendering.
	const GLhandleARB GetDebugRenderShader( )
	{ return (const GLhandleARB) m_hRenderPO; }

	/// \brief Returns the interface to the vertex position texture.
	const GLuint GetVtPosTex( )
	{ return (const GLuint) m_uiVtPosTexRT; };
	/// \brief Returns the interface to the texture coordinates texture.
	const GLuint GetVtTexCoordTex( )
	{ return (const GLuint) m_uiVtTexCoordTexRT; };
	/// \brief Returns the interface to the vertex normal texture.
	const GLuint GetVtNormalTex( )
	{ return (const GLuint) m_uiVtNormalTexRT; };
	/// \brief Returns the interface to the vertex tangent and handedness.
	const GLuint GetVtTanHTex( )
	{ return (const GLuint) m_uiVtTanHTexRT; };
	/// \brief Returns the interface to the 2nd tensor texture.
	const GLuint GetVtTensor2Tex( )
	{ return (const GLuint) m_uiVtTensor2TexRT; };
	/// \brief Returns the interface to the 3rd tensor texture.
	const GLuint GetVtTensor3Tex( )
	{ return (const GLuint) m_uiVtTensor3TexRT; };
	/// \brief Returns the interface to the #1 principal direction and k1.
	const GLuint GetVtP1K1Tex( )
	{ return (const GLuint) m_uiVtP1K1TexRT; };
	/// \brief Returns the interface to the#2 principal direction and k2.
	const GLuint GetVtP2K2Tex( )
	{ return (const GLuint) m_uiVtP2K2TexRT; };
	/// \brief Returns the current local transformation matrix
	const GLfloat* GetLocalTransform( ) { return m_mLocal; };
#endif

	int  SetVertexBuffer       ( float *pVtPos, float *pVtTexCoord, 
		                         float *pVtNormal, int iSize );
	int  SetIndexBuffer        ( int *pFaceIdx, int iSize );

	int  SetPreVertexDeform    ( char *szFunc /*, bool bDefTex*/ );
	int  SetPostVertexDeform   ( char *szFunc  );
	int  SetPixelDeform        ( char *szFunc  );
	void TogglePreVertexDeform ( bool bEnable  );
	void TogglePostVertexDeform( bool bEnable  );
	void TogglePixelDeform     ( bool bEnable  );
	/// \brief Returns whether vertex deformation before diff. geom estimation is enabled
	bool UsePreVertexDeform    ( ) { return m_bUsePreDefVS; };
	/// \brief Returns whether vertex deformation after diff. geom estimation is enabled
	bool UsePostVertexDeform   ( ) { return m_bUsePostDefVS; };
	/// \brief Returns whether pixel deformation is enabled
	bool UsePixelDeform        ( ) { return m_bUsePxDef; };

	void ToggleVisible         ( bool bVisible );
	/// \brief Returns whether the object is toggled for encoding.
	bool IsVisible             ( ) { return m_bVisible; };

	void SetUpdateCallback     ( void fUpdate( CIntObj *pObj, void *pUserData ) );
	void SetRenderCallback     ( void fIdx   ( CIntObj *pObj, int, void *pUserData ), 
							     void fNonIdx( CIntObj *pObj, void *pUserData ) );

	bool GetSemantic           ( INTSEMANTIC iIntSemantic, 
		                         VSSEMANTIC &iVsSemantic, int iIdx = 0 );
	void GetSemanticStr        ( VSSEMANTIC iVtSemantic, char *szResult );
	void GetParamStr           ( VSSEMANTIC iVtSemantic, char *szResult );
	void GetParamStrGLSL       ( VSSEMANTIC iVtSemantic, char *szResult );
	int  ReplaceString         ( int iNumOcc, char *szString, const char *szLookFor, 
							     const char *szReplace, char **szOut );

	/// \brief Sets the user-defined data used by the callback functions.
	/// \param pUserData Pointer to the user-defined data.
	void SetUserData       ( void *pUserData ) { m_pUserData = pUserData; };
	/// \brief Returns the user-defined data defined by CIntObj::SetUserData.
	const void* GetUserData( ) { return (const void*) m_pUserData; };

	CIntObj( CIntManager *pManager, const char *szAdjWriteFile = NULL, 
			 const char *szAdjReadFile = NULL );
	~CIntObj( );
};

#endif

////////////////////////////////////////////////////////////////////////////////
/// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
///
/// Permission is hereby granted, free of charge, to any person obtaining
/// a copy of this software and associated documentation files (the 
/// "Software"), to deal in the Software without restriction, including
/// without limitation the rights to use, copy, modify, merge, publish,
/// distribute, sublicense, and/or sell copies of the Software, and to
/// permit persons to whom the Software is furnished to do so, subject to
/// the following conditions:
///
/// The above copyright notice and this permission notice shall be
/// included in all copies or substantial portions of the Software.
///
/// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
/// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
/// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
/// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
/// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
/// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
/// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////////