////////////////////////////////////////////////////////////////////////////
// Fragment shader to compute per-face tangent and bitangent vectors
// with arbitrary vertex valency.
//
// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the 
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
// Size of a dimension of the vertex attribute map, in texels.
uniform int g_iVtMapSize;

uniform sampler2D s0; // Per-face vertex indices
uniform sampler2D s1; // Vertex position
uniform sampler2D s2; // Texture coordinates

////////////////////////////////////////////////////////////////////////////
// Convert 1D index (i) to 2D UV index.
// i = 1D index
// s = size of a screen line, in pixels
////////////////////////////////////////////////////////////////////////////
vec2 I1DTo2D(float i, float s) { return vec2(mod(i,s),int(i/s)); }

////////////////////////////////////////////////////////////////////////////
// Pixel shader used for computing face tangent bases.
////////////////////////////////////////////////////////////////////////////
void main( )
{
	// Sample face ID texture
	vec3 fVtIdx = vec3( texture2D( s0, gl_TexCoord[0].st ).xyz );
	if( fVtIdx.x == -1.0 ) discard;

	// Sample vertex position
	vec2 fVt1UV = I1DTo2D( fVtIdx.x, float(g_iVtMapSize) );
	vec2 fVt2UV = I1DTo2D( fVtIdx.y, float(g_iVtMapSize) );
	vec2 fVt3UV = I1DTo2D( fVtIdx.z, float(g_iVtMapSize) );	
    vec3 vV1 = vec3( texture2D( s1, fVt1UV ).xyz );
    vec3 vV2 = vec3( texture2D( s1, fVt2UV ).xyz );
    vec3 vV3 = vec3( texture2D( s1, fVt3UV ).xyz );

	vec3 vDV1 = vV2 - vV1;
	vec3 vDV2 = vV3 - vV1;
	
    // Face normal
	gl_FragData[0] = vec4( cross( vDV1, vDV2 ), 1.0 );

	// Sample texture coordinates
	vec2 vW1 = vec2( texture2D( s2, fVt1UV ).xy );
	vec2 vW2 = vec2( texture2D( s2, fVt2UV ).xy );
	vec2 vW3 = vec2( texture2D( s2, fVt3UV ).xy );

	vec2 vDW1 = vW2 - vW1;
	vec2 vDW2 = vW3 - vW1;

	float r = 1.0 / ( vDW1.x * vDW2.y - vDW2.x * vDW1.y );

	vec3 vS = vec3( vDW2.y * vDV1.x - vDW1.y * vDV2.x,
                    vDW2.y * vDV1.y - vDW1.y * vDV2.y,
                    vDW2.y * vDV1.z - vDW1.y * vDV2.z ) * r;
                        
	vec3 vT = vec3( vDW2.x * vDV1.x - vDW1.x * vDV2.x,
                    vDW2.x * vDV1.y - vDW1.x * vDV2.y,
                    vDW2.x * vDV1.z - vDW1.x * vDV2.z ) * r;
                      
	// ST tangents                                              
	gl_FragData[1] = vec4( vS, 1.0 );
	gl_FragData[2] = vec4( vT, 1.0 );	
}