////////////////////////////////////////////////////////////////////////////
// Shader effect to compute per-vertex normals with arbitrary 
// vertex valency.
//
// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the 
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
float2 g_fAdjFaceIdxC; // 1D-2D constant (face adjacency).
float2 g_fVtIdxC;      // 1D-2D constant (vertex attributes).
float2 g_fFaceIdxC;    // 1D-2D constant (per-face vertex indices).

////////////////////////////////////////////////////////////////////////////
// Vertex shader input semantics.
////////////////////////////////////////////////////////////////////////////
struct VS_IN
{
	float4 vPos : POSITION;  // Transformed position (XYZ+RHW).
    float2 vTex : TEXCOORD0; // Texture coordinates (UV).	
};

// Per-face vertex indices.
texture Tex0;
sampler s0 = sampler_state
{
    Texture   = (Tex0);	
    MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Vertex position.
texture Tex1;
sampler s1 = sampler_state
{
    Texture   = (Tex1);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Adjacency address.
texture Tex3;
sampler s3 = sampler_state
{
    Texture   = (Tex3);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Face adjacency address.
texture Tex4;
sampler s4 = sampler_state
{
    Texture   = (Tex4);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Face normal.
texture Tex5;
sampler s5 = sampler_state
{
    Texture   = (Tex5);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

////////////////////////////////////////////////////////////////////////////
// Convert 1D index (i) to 2D UV index.
// c = {1/W, 1/(W*H)} where W and H are the width and height of the
//     texture map
////////////////////////////////////////////////////////////////////////////
float2 I1DTo2D( float i, float2 c ) { return float2( i, i ) * c; } 

////////////////////////////////////////////////////////////////////////////
// Pixel shader used for computing face normals.
////////////////////////////////////////////////////////////////////////////
float4 FaceNPS( VS_IN In ) : COLOR
{
	// Sample face ID texture
	float3 fVtIdx = tex2D( s0, In.vTex ).xyz;
	if( fVtIdx.x == -1.0 ) discard;
	
	// Sample vertex position
	float2 fVt1UV = I1DTo2D( fVtIdx.x, g_fVtIdxC );
	float2 fVt2UV = I1DTo2D( fVtIdx.y, g_fVtIdxC );
	float2 fVt3UV = I1DTo2D( fVtIdx.z, g_fVtIdxC );	
    float3 vV1 = tex2D( s1, fVt1UV ).xyz;
    float3 vV2 = tex2D( s1, fVt2UV ).xyz;
    float3 vV3 = tex2D( s1, fVt3UV ).xyz;
    
	float3 vDV1 = vV2 - vV1;
	float3 vDV2 = vV3 - vV1;
	
	return float4( cross( vDV1, vDV2 ), 1.0 );
}

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing vertex normals.
////////////////////////////////////////////////////////////////////////////
float4 VertexNPS( VS_IN In ) : COLOR
{
	float3 vNormal = float3( 0.0, 0.0, 0.0 );
	
	// Sample adjacency addresses
	// .x) Address to the 1st adjacent face in the face adjacency texture
	// .y) Number of adjacent faces
	// .z) Address to the 1st adjacent vertex in the vertex adjacency 
	//     texture
	// .w) Number of adjacent vertices
	float4 fAdj = tex2D( s3,In.vTex );
	if( fAdj.y == -1.0 ) discard;
	
	// This vertex
	float3 vV = tex2D( s1, In.vTex );
	
	int iCt = 0;
	while( iCt < fAdj.y )
    {
		// Get face index in the face adjacency texture
		float2 vAdjUV  = I1DTo2D( fAdj.x + iCt, g_fAdjFaceIdxC );		
		float fFaceIdx = tex2Dlod( s4, float4( vAdjUV,0,0 ) ).x;

		// Get face normal
		float2 vFaceUV     = I1DTo2D( fFaceIdx, g_fFaceIdxC );		
		float3 vFaceNormal = tex2Dlod( s5, float4( vFaceUV,0,0 ) ).xyz;
		
		// Add
		vNormal += vFaceNormal;
		
		iCt++;
    } 
    
	return float4( normalize( vNormal ), 1.0f ); 
}

////////////////////////////////////////////////////////////////////////////
// Effect technique for computing face normals.
////////////////////////////////////////////////////////////////////////////
technique FaceN
{
    pass P0
    {          
        VertexShader = null;
        PixelShader  = compile ps_2_0 FaceNPS( );
    }
}

////////////////////////////////////////////////////////////////////////////
// Effect technique for computing vertex normals.
////////////////////////////////////////////////////////////////////////////
technique VertexN
{
    pass P0
    {          
        VertexShader = null;
        PixelShader  = compile ps_3_0 VertexNPS( );
    }
}