////////////////////////////////////////////////////////////////////////////
// Shader effect to compute per-vertex elements of differential geometry 
// of 2nd and 3rd order.
//
// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the 
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
float2 g_fAdjVtIdxC; // 1D-2D constant (vertex adjacency).
float2 g_fVtIdxC;    // 1D-2D constant (vertex attributes).

////////////////////////////////////////////////////////////////////////////
// Vertex shader input semantics.
////////////////////////////////////////////////////////////////////////////
struct VS_IN
{
	float4 vPos : POSITION;  // Transformed position (XYZ+RHW).
    float2 vTex : TEXCOORD0; // Texture coordinates (UV).		
};

////////////////////////////////////////////////////////////////////////////
// Pixel shader output semantics for Tensor2.
////////////////////////////////////////////////////////////////////////////
struct PS_OUT1
{
	// 0: 2nd fundamental form (3 values)
	// 1: pdir1, k1 (4 values)
	// 2: pdir2, k2 (4 values)
    float4 vColor[3] : COLOR;
};

////////////////////////////////////////////////////////////////////////////
// Pixel shader output semantics for Tensor3.
////////////////////////////////////////////////////////////////////////////
struct PS_OUT2
{
	// 3rd fundamental form (4 values)
    float4 vColor : COLOR;
};

// Vertex position.
texture Tex0;
sampler s0 = sampler_state
{
    Texture   = (Tex0);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Per-vertex normal.
texture Tex1;
sampler s1 = sampler_state
{
    Texture   = (Tex1);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Adjacency address.
texture Tex2;
sampler s2 = sampler_state
{
    Texture   = (Tex2);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Vertex adjacency.
texture Tex3;
sampler s3 = sampler_state
{
    Texture   = (Tex3);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Principal direction and curvature k1.
texture Tex4;
sampler s4 = sampler_state
{
    Texture   = (Tex4);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Principal direction and curvature k2.
texture Tex5;
sampler s5 = sampler_state
{
    Texture   = (Tex5);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

#define M_2_PI 0.636619772367581343076
#define M_PI_2 1.57079632679489661923
#define M_PI 3.14159265

float signal( float x )
{
   if( x <  0 ) 
       return -1; 
   else
       if( x == 0 ) 
           return 0; 
       else 
           return 1;
}

float3 hsv(float3 hsv)
{
	// From FvD
	hsv.x = fmod(hsv.x, 2.0 * M_PI);
	if (hsv.x < 0.0)
		hsv.x += 2.0 * M_PI;
	hsv.x /= M_PI / 3.0;
	int i = int(floor(hsv.x));
	double f = hsv.x - i;
	double p = hsv.z * (1.0 - hsv.y);
	double q = hsv.z * (1.0 - (hsv.y*f));
	double t = hsv.z * (1.0 - (hsv.y*(1.0-f)));
	
	if( i == 0 ) return float3(hsv.z, t, p);
	else if( i == 1 ) return float3(q, hsv.z, p);
	else if( i == 2 ) return float3(p, hsv.z, t);
	else if( i == 3 ) return float3(p, q, hsv.z);
	else if( i == 4 ) return float3(t, p, hsv.z);
	else return float3(hsv.z, p, q);
}

////////////////////////////////////////////////////////////////////////////
// Perform LDL^T decomposition of a 3x3 symmetric positive definite matrix
// and solves Ax=B.
////////////////////////////////////////////////////////////////////////////
float3 ldltsv3( float3x3 w, float3 m )
{
	// Decompose matrix. Overwrites lower triangle of matrix.
	float3 d;
	float2 v;
	int i;
	for( i = 0; i < 3; i++ ) 
	{
		for( int k = 0; k < i; k++ ) v[k] = w[i][k] * d[k];
		for( int j = i; j < 3; j++ ) 
		{
			float sum = w[i][j];
			for( int k = 0; k < i; k++ ) 
				sum -= v[k] * w[j][k];
			if( i == j ) 
				d[i] = 1.0f / sum;
			else
				w[j][i] = sum;
		}
	}
	
	// Solve Ax=B after LDL^T decomposition
	for( i = 0; i < 3; i++ )
	{
		float sum = m[i];
		for( int k = 0; k < i; k++ ) 
			sum -= w[i][k] * m[k];
		m[i] = sum * d[i];
	}
	for( i = 2; i >= 0; i-- )
	{
		float sum = 0;
		for( int k = i + 1; k < 3; k++ ) 
			sum += w[k][i] * m[k];
		m[i] -= sum * d[i];
	}
	
	return m;
}

////////////////////////////////////////////////////////////////////////////
// Perform LDL^T decomposition of a 4x4 symmetric positive definite matrix
// and solves Ax=B.
////////////////////////////////////////////////////////////////////////////
float4 ldltsv4( float4x4 w, float4 m )
{
	// Decompose matrix. Overwrites lower triangle of matrix.
	float4 d;
	float3 v;
	int i;
	for( i = 0; i < 4; i++ ) 
	{
		for( int k = 0; k < i; k++ ) v[k] = w[i][k] * d[k];
		for( int j = i; j < 4; j++ ) 
		{
			float sum = w[i][j];
			for( int k = 0; k < i; k++ ) 
				sum -= v[k] * w[j][k];
			if( i == j ) 
				d[i] = 1.0f / sum;
			else
				w[j][i] = sum;
		}
	}
	
	// Solve Ax=B after LDL^T decomposition
	for( i = 0; i < 4; i++ )
	{
		float sum = m[i];
		for( int k = 0; k < i; k++ ) 
			sum -= w[i][k] * m[k];
		m[i] = sum * d[i];
	}
	for( i = 3; i >= 0; i-- )
	{
		float sum = 0;
		for( int k = i + 1; k < 4; k++ ) 
			sum += w[k][i] * m[k];
		m[i] -= sum * d[i];
	}
	
	return m;
}

////////////////////////////////////////////////////////////////////////////
// Convert 1D index (i) to 2D UV index.
// c = {1/W, 1/(W*H)} where W and H are the width and height of the
//     texture map
////////////////////////////////////////////////////////////////////////////
float2 I1DTo2D( float i, float2 c ) { return float2( i, i ) * c; } 

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing 2nd fundamental forms.
////////////////////////////////////////////////////////////////////////////
PS_OUT1 Tensor2PS( VS_IN In )
{
	PS_OUT1 Out;
	
	// Sample adjacency addresses
	// .x) Address to the 1st adjacent face in the face adjacency texture
	// .y) Number of adjacent faces
	// .z) Address to the 1st adjacent vertex in the vertex adjacency 
	//     texture
	// .w) Number of adjacent vertices
	float4 fAdj = tex2D( s2, In.vTex );
	if( fAdj.z == -1.0 ) discard;
	
	// UV index to the first adjacent vertex
	float2 vAdjUVNei  = I1DTo2D( fAdj.z, g_fAdjVtIdxC );
	float  fIdxNei    = tex2D( s3, vAdjUVNei ).x;	
	float2 vUVNei     = I1DTo2D( fIdxNei, g_fVtIdxC );
	
	float3 vPos       = tex2D( s0, In.vTex );
	float3 vNormal    = tex2D( s1, In.vTex ); 
	float3 vPosNei    = tex2D( s0, vUVNei  );
	float3 vNormalNei = tex2D( s1, vUVNei  );
	float3 vPDir1     = normalize( cross( (vPosNei - vPos), vNormal ) );
	float3 vPDir2     = cross( vNormal, vPDir1 );
	
	float3 m = 0;	
	float3x3 w = 0;
	
	int iCt = 0;
	while( iCt < fAdj.w )
    {
		// Get neighbor index from the vertex adjacency texture
		vAdjUVNei  = I1DTo2D( fAdj.z + iCt, g_fAdjVtIdxC );		
		fIdxNei    = tex2Dlod( s3, float4( vAdjUVNei,0,0 ) ).x;
		
		// Get attributes of neighbor vertex		
		vUVNei     = I1DTo2D( fIdxNei, g_fVtIdxC );		
		vPosNei    = tex2Dlod( s0, float4( vUVNei,0,0 ) );
		vNormalNei = tex2Dlod( s1, float4( vUVNei,0,0 ) );		
    
		float3 e = vPosNei - vPos;
		float  u = dot( e, vPDir1 );
		float  v = dot( e, vPDir2 );
		w[0][0] += u*u;
		w[0][1] += u*v;
		w[2][2] += v*v;
		float3 dn = vNormalNei - vNormal;
		float dnu = dot( dn, vPDir1 );
		float dnv = dot( dn, vPDir2 );
		m[0] += dnu*u;
		m[1] += dnu*v + dnv*u;
		m[2] += dnv*v;
	
		iCt++;		
    } 
    
    w[1][1] = w[0][0] + w[2][2];
    w[1][2] = w[0][1];
    
	m = ldltsv3( w, m );
	
	// 2nd fundamental form
	Out.vColor[0] = float4( m, 0 );
	
	// Jacobi rotation to diagonalize
	float ku = m[0], kuv = m[1], kv = m[2];
	float c = 1, s = 0, tt = 0;
	if( kuv != 0.0f ) 
	{
		float h = 0.5f * (kv - ku) / kuv;
		tt = (h < 0.0f) ? 1.0f / (h - (float) sqrt(1.0f + h*h)) :
			 1.0f / (h + (float) sqrt(1.0f + h*h));
		c = 1.0f / (float) sqrt(1.0f + tt*tt);
		s = tt * c;
	}

	float fK1 = ku - tt * kuv, 
		  fK2 = kv + tt * kuv;

	float3 pd1, pd2;
	if( abs( fK1 ) >= abs( fK2 ) )
		pd1 = c * vPDir1 - s * vPDir2;
	else 
	{
		float fTemp = fK1; fK1 = fK2; fK1 = fTemp;
		pd1 = s * vPDir1 + c * vPDir2;
	}
	pd2 = cross( vNormal, pd1 );

	// Principal direction of k1
	Out.vColor[1] = float4( pd1,fK1 );
	
	//
/*
	float cscale = 1.0f;
	float H = 0.5f * (fK1 + fK2);
	float K = fK1 * fK2;
	
	float h = 4.0f / 3.0f * (float) abs(atan2(H*H-K,H*H*signal(H)));
	float sc = (float) M_2_PI * (float) atan((2.0f*H*H-K)*cscale);
	
	float3 teste = float3(h,sc,1.0);
	if ( teste.y <= 0.0)
		Out.vColor[1] = float4( teste.z, teste.z, teste.z, 1.0f );	
	else	
		Out.vColor[1] = float4( hsv(teste), 1.0f );
*/
//	float teste = int(min(max(256.0 / (float) M_PI * (atan(H*cscale) + M_PI_2),
//			    0.0), 255.99));
//	Out.vColor[1] = float4(teste / 255, teste / 255, teste / 255,1.0f);

//

	// Principal direction of k2
	Out.vColor[2] = float4( pd2,fK2 );
	
	return Out;
}

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing 3rd fundamental forms.
////////////////////////////////////////////////////////////////////////////
PS_OUT2 Tensor3PS( VS_IN In )
{
	PS_OUT2 Out;
	
	// Sample adjacency addresses
	// .x) Address to the 1st adjacent face in the face adjacency texture
	// .y) Number of adjacent faces
	// .z) Address to the 1st adjacent vertex in the vertex adjacency 
	//     texture
	// .w) Number of adjacent vertices
	float4 fAdj = tex2D( s2, In.vTex );
	if( fAdj.z == -1.0 ) discard;
	
	float3 vPos   = tex2D( s0, In.vTex );	
	float4 vpk1   = tex2D( s4, In.vTex ); 
	float4 vpk2   = tex2D( s5, In.vTex ); 	

	float3 vPDir1 = vpk1.xyz, 
	       vPDir2 = vpk2.xyz;
	float2 vCurv  = float2( vpk1.w, vpk2.w );
	float2 vCurvNei;

	float4 m = 0;	
	float4x4 w = 0;
	
	int iCt = 0;
	while( iCt < fAdj.w )
    {
		// Get neighbor index from the vertex adjacency texture
		float2 vAdjUVNei = I1DTo2D( fAdj.z + iCt, g_fAdjVtIdxC );		
		float  fIdxNei   = tex2Dlod( s3, float4( vAdjUVNei,0,0 ) ).x;
		
		// Get attributes of neighbor vertex		
		float2 vUVNei    = I1DTo2D( fIdxNei, g_fVtIdxC );		
		float3 vPosNei   = tex2Dlod( s0, float4( vUVNei,0,0 ) );
		vCurvNei.x       = tex2Dlod( s4, float4( vUVNei,0,0 ) ).w;
		vCurvNei.y       = tex2Dlod( s5, float4( vUVNei,0,0 ) ).w;
    
		float3 fcurv_i = float3( vCurv.x,    0, vCurv.y    ), 
		       fcurv_n = float3( vCurvNei.x, 0, vCurvNei.y );
    
		// Variation of curvature along each edge
		float3 dfcurv = fcurv_n - fcurv_i;
		float3 e = vPosNei - vPos;

		float u = dot( e, vPDir1 );
		float v = dot( e, vPDir2 );
		float u2 = u*u, v2 = v*v, uv = u*v;
		w[0][0] += u2;
		w[0][1] += uv;
		w[3][3] += v2;
		m[0] += u*dfcurv[0];
		m[1] += v*dfcurv[0] + 2.0f*u*dfcurv[1];
		m[2] += 2.0f*v*dfcurv[1] + u*dfcurv[2];
		m[3] += v*dfcurv[2];
    	
		iCt++;
    } 
    
	w[1][1] = 2.0f * w[0][0] + w[3][3];
	w[1][2] = 2.0f * w[0][1];
	w[2][2] = w[0][0] + 2.0f * w[3][3];
	w[2][3] = w[0][1];
    
	m = ldltsv4( w, m );
	
	// 3rd fundamental form
	Out.vColor = m;
/*
	float cscale = 1.0f;
	float H = 0.25f * (m[0] + m[1] + m[2] + m[3]);
	float K = m[0] * m[1] * m[2] * m[3];
	
	float h = 4.0f / 3.0f * (float) abs(atan2(H*H-K,H*H*signal(H)));
	float sc = (float) M_2_PI * (float) atan((2.0f*H*H-K)*cscale);
	
	float3 teste = float3(h,sc,1.0);
	if ( teste.y <= 0.0)
		Out.vColor = float4( teste.z, teste.z, teste.z, 1.0f );	
	else	
		Out.vColor = float4( hsv(teste), 1.0f );
		

	float teste = int(min(max(256.0 / (float) M_PI * (atan(H*cscale) + M_PI_2),
			    0.0), 255.99));
	Out.vColor = float4(teste / 255, teste / 255, teste / 255,1.0f);
*/
	return Out;
}

////////////////////////////////////////////////////////////////////////////
// Effect technique for computing per-vertex 2nd order tensor
////////////////////////////////////////////////////////////////////////////
technique Tensor2
{
    pass P0
    {          
        VertexShader = null;
        PixelShader  = compile ps_3_0 Tensor2PS( );
    }
}

////////////////////////////////////////////////////////////////////////////
// Effect technique for computing per-vertex 3rd order tensor
////////////////////////////////////////////////////////////////////////////
technique Tensor3
{
    pass P0
    {          
        VertexShader = null;
        PixelShader  = compile ps_3_0 Tensor3PS( );
    }
}