////////////////////////////////////////////////////////////////////////////
// Fragment shader to compute per-vertex elements of differential geometry 
// of 2nd order.
//
// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the 
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
int g_iAdjVtMapSize;  // Size of vertex adjacency map
int g_iVtMapSize;     // Size of the vertex attribute map

uniform sampler2D s0; // Vertex position
uniform sampler2D s1; // Per-vertex normal
uniform sampler2D s2; // Adjacency address
uniform sampler2D s3; // Vertex adjacency

const float M_2_PI = 0.636619772367581343076;
const float M_PI_2 = 1.57079632679489661923;
const float M_PI = 3.14159265;

vec3 hsv(vec3 hsv)
{
	// From FvD
	hsv.x = mod(hsv.x, 2.0 * M_PI);
	if (hsv.x < 0.0)
		hsv.x += 2.0 * M_PI;
	hsv.x /= M_PI / 3.0;
	int i = int(floor(hsv.x));
	float f = hsv.x - float(i);
	float p = hsv.z * (1.0 - hsv.y);
	float q = hsv.z * (1.0 - (hsv.y*f));
	float t = hsv.z * (1.0 - (hsv.y*(1.0-f)));
	
	if( i == 0 ) return vec3(hsv.z, t, p);
	else if( i == 1 ) return vec3(q, hsv.z, p);
	else if( i == 2 ) return vec3(p, hsv.z, t);
	else if( i == 3 ) return vec3(p, q, hsv.z);
	else if( i == 4 ) return vec3(t, p, hsv.z);
	else return vec3(hsv.z, p, q);
}

////////////////////////////////////////////////////////////////////////////
// Perform LDL^T decomposition of a 3x3 symmetric positive definite matrix
// and solves Ax=B.
////////////////////////////////////////////////////////////////////////////
vec3 ldltsv3( mat3 w, vec3 m )
{
	// Decompose matrix. Overwrites lower triangle of matrix.
	vec3 d;
	vec2 v;
	int i;
	for( i = 0; i < 3; i++ ) 
	{
		for( int k = 0; k < i; k++ ) v[k] = w[i][k] * d[k];
		for( int j = i; j < 3; j++ ) 
		{
			float sum = w[i][j];
			for( int k = 0; k < i; k++ ) 
				sum -= v[k] * w[j][k];
			if( i == j ) 
				d[i] = 1.0 / sum;
			else
				w[j][i] = sum;
		}
	}
	
	// Solve Ax=B after LDL^T decomposition
	for( i = 0; i < 3; i++ )
	{
		float sum = m[i];
		for( int k = 0; k < i; k++ ) 
			sum -= w[i][k] * m[k];
		m[i] = sum * d[i];
	}
	for( i = 2; i >= 0; i-- )
	{
		float sum = 0.0;
		for( int k = i + 1; k < 3; k++ ) 
			sum += w[k][i] * m[k];
		m[i] -= sum * d[i];
	}
	
	return m;
}


////////////////////////////////////////////////////////////////////////////
// Convert 1D index (i) to 2D UV index.
// i = 1D index
// s = size of a screen line, in pixels
////////////////////////////////////////////////////////////////////////////
vec2 I1DTo2D(float i, float s) { return vec2(mod(i,s),int(i/s)); }

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing 2nd fundamental forms.
////////////////////////////////////////////////////////////////////////////
void main( )
{
	// Sample adjacency addresses
	// .x) Address to the 1st adjacent face in the face adjacency texture
	// .y) Number of adjacent faces
	// .z) Address to the 1st adjacent vertex in the vertex adjacency 
	//     texture
	// .w) Number of adjacent vertices
	vec4 fAdj = texture2D( s2, gl_TexCoord[0].st );
	if( fAdj.z == -1.0 ) discard;

	// UV index to the first adjacent vertex
	vec2  vAdjUVNei  = I1DTo2D( fAdj.z, float(g_iAdjVtMapSize) );
	float fIdxNei    = texture2D( s3, vAdjUVNei ).x;	
	vec2  vUVNei     = I1DTo2D( fIdxNei, float(g_iVtMapSize) );
	
	vec3  vPos       = texture2D( s0, gl_TexCoord[0].st ).xyz;
	vec3  vNormal    = texture2D( s1, gl_TexCoord[0].st ).xyz; 
	vec3  vPosNei    = texture2D( s0, vUVNei ).xyz;
	vec3  vNormalNei = texture2D( s1, vUVNei ).xyz;
	vec3  vPDir1     = normalize( cross( (vPosNei - vPos), vNormal ) );
	vec3  vPDir2     = cross( vNormal, vPDir1 );
	
	vec3 m = vec3( 0.0 );	
	mat3 w = mat3( 0.0 );
	
	float fCt = 0.0;
	while( fCt < fAdj.w )
    {
		// Get neighbor index from the vertex adjacency texture
		vAdjUVNei  = I1DTo2D( fAdj.z + fCt, float(g_iAdjVtMapSize) );
		fIdxNei    = texture2D( s3, vAdjUVNei ).x;
		
		// Get attributes of neighbor vertex		
		vUVNei     = I1DTo2D( fIdxNei, float(g_iVtMapSize) );
		vPosNei    = texture2D( s0, vUVNei ).xyz;
		vNormalNei = texture2D( s1, vUVNei ).xyz;		
    
		vec3   e = vPosNei - vPos;
		float  u = dot( e, vPDir1 );
		float  v = dot( e, vPDir2 );
		w[0][0] += u*u;
		w[0][1] += u*v;
		w[2][2] += v*v;
		vec3 dn  = vNormalNei - vNormal;
		float dnu = dot( dn, vPDir1 );
		float dnv = dot( dn, vPDir2 );
		m[0] += dnu*u;
		m[1] += dnu*v + dnv*u;
		m[2] += dnv*v;
	
		fCt++;		
    } 
    
    w[1][1] = w[0][0] + w[2][2];
    w[1][2] = w[0][1];
    
	m = ldltsv3( w, m );
	
	// 2nd fundamental form
	gl_FragData[0] = vec4( m, 0 );
	
	// Jacobi rotation to diagonalize
	float ku = m[0], kuv = m[1], kv = m[2];
	float c = 1.0, s = 0.0, tt = 0.0;
	if( kuv != 0.0 ) 
	{
		float h = 0.5 * (kv - ku) / kuv;
		tt = (h < 0.0) ? 1.0 / (h - sqrt(1.0 + h*h)) : 1.0 / (h + sqrt(1.0 + h*h));
		c = 1.0 / sqrt(1.0 + tt*tt);
		s = tt * c;
	}

	float fK1 = ku - tt * kuv, 
		  fK2 = kv + tt * kuv;

	vec3 pd1, pd2;
	if( abs( fK1 ) >= abs( fK2 ) )
		pd1 = c * vPDir1 - s * vPDir2;
	else 
	{
		float fTemp = fK1; fK1 = fK2; fK1 = fTemp;
		pd1 = s * vPDir1 + c * vPDir2;
	}
	pd2 = cross( vNormal, pd1 );

	// Principal direction of k1
	gl_FragData[1] = vec4( pd1,fK1 );
	
	//
/*
	float cscale = 1.0;
	float H = 0.5 * (fK1 + fK2);
	float K = fK1 * fK2;
	
	float h = 4.0 / 3.0 * abs(atan((H*H-K) / (H*H*sign(H))));
	float sc = M_2_PI * atan((2.0f*H*H-K)*cscale);
	
	vec3 teste = float3(h,sc,1.0);
	if ( teste.y <= 0.0)
		gl_FragData[1] = vec4( teste.z, teste.z, teste.z, 1.0f );	
	else	
		gl_FragData[1] = vec4( hsv(teste), 1.0f );
*/
//	float teste = int(min(max(256.0 / M_PI * (atan(H*cscale) + M_PI_2),
//			    0.0), 255.99));
//	gl_FragData[1] = vec4(teste / 255, teste / 255, teste / 255, 1.0);

//

	// Principal direction of k2
	gl_FragData[2] = vec4( pd2,fK2 );
}
