////////////////////////////////////////////////////////////////////////////
// Vertex shader to compute per-vertex tangent and bitangent vectors
// with arbitrary vertex valency.
//
// Copyright (C) 2007 Harlen Costa Batagelo. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the 
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
//
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
uniform int g_iAdjFaceMapSize; // Size of face adjacency map
uniform int g_iFaceMapSize;    // Size of per-face vertex indices map

uniform sampler2D s1; // Vertex position
uniform sampler2D s3; // Adjacency address
uniform sampler2D s4; // Face adjacency address
uniform sampler2D s5; // Face normal
uniform sampler2D s6; // Face tangent S
uniform sampler2D s7; // Face tangent T

////////////////////////////////////////////////////////////////////////////
// Convert 1D index (i) to 2D UV index.
// i = 1D index
// s = size of a screen line, in pixels
////////////////////////////////////////////////////////////////////////////
vec2 I1DTo2D(float i, float s) { return vec2(mod(i,s),int(i/s)); }

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing vertex tangent bases.
////////////////////////////////////////////////////////////////////////////
void main( )
{
	vec3 vNormal, vS, vT;
	vNormal = vS = vT = vec3( 0.0, 0.0, 0.0 );
	
	// Sample adjacency addresses
	// .x) Address to the 1st adjacent face in the face adjacency texture
	// .y) Number of adjacent faces
	// .z) Address to the 1st adjacent vertex in the vertex adjacency 
	//     texture
	// .w) Number of adjacent vertices
	vec4 fAdj = texture2D( s3, gl_TexCoord[0].st );
	if( fAdj.y == -1.0 ) discard;
	
	float fCt = 0.0;
	while( fCt < fAdj.y )
    {
		// Get face index in the face adjacency texture
		vec2 vAdjUV  = I1DTo2D( fAdj.x + fCt, float(g_iAdjFaceMapSize) );
		float fFaceIdx = texture2D( s4, vAdjUV ).x;

		// Get face normal
		vec2 vFaceUV     = I1DTo2D( fFaceIdx, float(g_iFaceMapSize) );
		vec3 vFaceNormal = texture2D( s5, vFaceUV ).xyz;
		vec3 vFaceTanS   = texture2D( s6, vFaceUV ).xyz;
		vec3 vFaceTanT   = texture2D( s7, vFaceUV ).xyz;

		// Add
		vNormal += vFaceNormal;
		vS      += vFaceTanS;
		vT      += vFaceTanT;
		
		fCt++;
    } 
    
	// Vertex normal
	vec3 vVN = normalize( vNormal );
	gl_FragData[0] = vec4( vVN, 1.0 ); 

	// Gram-Schmidt orthogonalize
	vec4 vTan = vec4( normalize( vS - vVN * dot( vVN, vS ) ), 1.0 );

	// Calculate handedness
	if( dot( cross( vVN, vS ), vT ) < 0.0 )
		vTan.w = -1.0;
    
	gl_FragData[1] = vTan;
}
